

import logging
import os

from rcsb.utils.config.ConfigUtil import ConfigUtil

logging.basicConfig(level=logging.INFO, format='%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s')
logger = logging.getLogger()
logger.setLevel(logging.INFO)

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))


class MakeConfig(object):

    def __init__(self):
        #
        self.__inpPathConfigIni = os.path.join(HERE, 'dbload-setup-example.cfg')
        cfgOb = ConfigUtil(configPath=self.__inpPathConfigIni, mockTopPath=os.path.join(TOPDIR, 'rcsb', 'mock-data'))
        oD = cfgOb.exportConfig()
        for ky in oD:
            logger.info("section name %s" % ky)
            for k, v in oD[ky].items():
                logger.info("op:  %s val: %s" % (k, v))

        self.__outPathConfigYamlExport1 = os.path.join(TOPDIR, 'rcsb', 'mock-data', 'config', 'dbload-setup-example.yml')
        self.__outPathConfigYamlExport2 = os.path.join(TOPDIR, 'rcsb', 'mock-data', 'config', 'dbload-setup-example-copy.yml')
        #cfgOb = ConfigUtil(configPath=self.__inpPathConfigIni, mockTopPath=os.path.join(TOPDIR, 'rcsb', 'mock-data'))
        cfgOb = ConfigUtil(mockTopPath=os.path.join(TOPDIR, 'rcsb', 'mock-data'))
        #
        cD = self.__setup()
        #
        oD.update(cD)
        #
        cfgOb.importConfig(oD)
        #
        ok = cfgOb.writeConfig(self.__outPathConfigYamlExport1, configFormat='yaml')
        logger.info("status %r" % ok)
        cfgOb = ConfigUtil(configPath=self.__outPathConfigYamlExport1, configFormat='yaml', mockTopPath=os.path.join(TOPDIR, 'rcsb', 'mock-data'))
        cfgOb.exportConfig()
        #logger.info("recovered config %r" % rD['dictionary_helper']['selection_filters'])
        #
        ok = cfgOb.writeConfig(self.__outPathConfigYamlExport2, configFormat='yaml')

    def __setup(self):
        dH = {}
        sD = {}
        dD = {}
        cfgD = {'dictionary_helper': dH, 'schemadef_helper': sD, 'document_helper': dD}
        #
        # Data items requiring a particular data transformation -
        dH['item_transformers'] = {
            'STRIP_WS': [
                {'CATEGORY_NAME': 'entity_poly', 'ATTRIBUTE_NAME': 'pdbx_c_terminal_seq_one_letter_code'},
                {'CATEGORY_NAME': 'entity_poly', 'ATTRIBUTE_NAME': 'pdbx_n_terminal_seq_one_letter_code'},
                {'CATEGORY_NAME': 'entity_poly', 'ATTRIBUTE_NAME': 'pdbx_seq_one_letter_code'},
                {'CATEGORY_NAME': 'entity_poly', 'ATTRIBUTE_NAME': 'pdbx_seq_one_letter_code_can'},
                {'CATEGORY_NAME': 'entity_poly', 'ATTRIBUTE_NAME': 'pdbx_seq_one_letter_code_sample'},
                {'CATEGORY_NAME': 'struct_ref', 'ATTRIBUTE_NAME': 'pdbx_seq_one_letter_code'}
            ]
        }

        dH['cardinality_parent_items'] = {
            'bird': {'CATEGORY_NAME': 'pdbx_reference_molecule', 'ATTRIBUTE_NAME': 'prd_id'},
            'bird_family': {'CATEGORY_NAME': 'pdbx_reference_molecule_family', 'ATTRIBUTE_NAME': 'family_prd_id'},
            'chem_comp': {'CATEGORY_NAME': 'chem_comp', 'ATTRIBUTE_NAME': 'id'},
            'bird_chem_comp': {'CATEGORY_NAME': 'chem_comp', 'ATTRIBUTE_NAME': 'id'},
            'pdbx': {'CATEGORY_NAME': 'entry', 'ATTRIBUTE_NAME': 'id'},
            'pdbx_core': {'CATEGORY_NAME': 'entry', 'ATTRIBUTE_NAME': 'id'}
        }
        dH['cardinality_category_extras'] = ['rcsb_load_status']
        #
        dH['selection_filters'] = {('PUBLIC_RELEASE', 'pdbx'): [{'CATEGORY_NAME': 'pdbx_database_status', 'ATTRIBUTE_NAME': 'status_code', 'VALUES': ['REL']}],
                                   ('PUBLIC_RELEASE', 'pdbx_core'): [{'CATEGORY_NAME': 'pdbx_database_status', 'ATTRIBUTE_NAME': 'status_code', 'VALUES': ['REL']}],
                                   ('PUBLIC_RELEASE', 'chem_comp'): [{'CATEGORY_NAME': 'chem_comp', 'ATTRIBUTE_NAME': 'pdbx_release_status', 'VALUES': ['REL', 'OBS', 'REF_ONLY']}],
                                   ('PUBLIC_RELEASE', 'bird_chem_comp'): [{'CATEGORY_NAME': 'chem_comp', 'ATTRIBUTE_NAME': 'pdbx_release_status', 'VALUES': ['REL', 'OBS', 'REF_ONLY']}],
                                   ('PUBLIC_RELEASE', 'bird'): [{'CATEGORY_NAME': 'pdbx_reference_molecule', 'ATTRIBUTE_NAME': 'release_status', 'VALUES': ['REL', 'OBS']}],
                                   ('PUBLIC_RELEASE', 'bird_family'): [{'CATEGORY_NAME': 'pdbx_reference_molecule_family', 'ATTRIBUTE_NAME': 'release_status', 'VALUES': ['REL', 'OBS']}]
                                   }
        #
        dH['type_code_classes'] = {'iterable': ['ucode-alphanum-csv', 'id_list', 'alphanum-scsv', 'alphanum-csv']}
        dH['query_string_selectors'] = {'iterable': ['comma separate']}
        # Put the non default iterable delimiter cases here -
        dH['iterable_delimiters'] = [{'CATEGORY_NAME': 'chem_comp', 'ATTRIBUTE_NAME': 'pdbx_synonyms', 'DELIMITER': ';'},
                                     {'CATEGORY_NAME': 'citation', 'ATTRIBUTE_NAME': 'rcsb_authors', 'DELIMITER': ';'}]
        #
        # Categories/Attributes that will be included in a schema definitions even if they are not populated in any tabulated instance data -
        #
        #
        #
        dH['content_classes'] = {('GENERATED_CONTENT', 'pdbx'): [{'CATEGORY_NAME': 'rcsb_load_status', 'ATTRIBUTE_NAME_LIST': ['datablock_name', 'load_date', 'locator']},
                                                                 {'CATEGORY_NAME': 'pdbx_struct_assembly_gen', 'ATTRIBUTE_NAME_LIST': ['ordinal']}],
                                 ('GENERATED_CONTENT', 'pdbx_core'): [{'CATEGORY_NAME': 'rcsb_load_status', 'ATTRIBUTE_NAME_LIST': ['datablock_name', 'load_date', 'locator']},
                                                                      {'CATEGORY_NAME': 'citation', 'ATTRIBUTE_NAME_LIST': ['rcsb_authors']},
                                                                      {'CATEGORY_NAME': 'pdbx_struct_assembly_gen', 'ATTRIBUTE_NAME_LIST': ['ordinal']},
                                                                      {'CATEGORY_NAME': 'pdbx_struct_assembly', 'ATTRIBUTE_NAME_LIST': ['rcsb_details', 'rcsb_candidate_assembly']},
                                                                      {'CATEGORY_NAME': 'rcsb_entry_container_identifiers', 'ATTRIBUTE_NAME_LIST': [
                                                                          'entry_id', 'entity_ids', 'polymer_entity_ids', 'non-polymer_entity_ids', 'assembly_ids']},
                                                                      {'CATEGORY_NAME': 'rcsb_entity_container_identifiers', 'ATTRIBUTE_NAME_LIST': [
                                                                          'entry_id', 'entity_id', 'asym_ids', 'auth_asym_ids']},
                                                                      {'CATEGORY_NAME': 'rcsb_assembly_container_identifiers', 'ATTRIBUTE_NAME_LIST': ['entry_id', 'assembly_id']},
                                                                      {'CATEGORY_NAME': 'rcsb_entity_source_organism', 'ATTRIBUTE_NAME_LIST': ['entity_id', 'pdbx_src_id', 'source_type',
                                                                                                                                               'scientific_name', 'common_name', 'ncbi_taxonomy_id',
                                                                                                                                               'provenance_code', 'beg_seq_num', 'end_seq_num']},
                                                                      {'CATEGORY_NAME': 'rcsb_entity_host_organism', 'ATTRIBUTE_NAME_LIST': ['entity_id', 'pdbx_src_id',
                                                                                                                                             'scientific_name', 'common_name', 'ncbi_taxonomy_id',
                                                                                                                                             'provenance_code', 'beg_seq_num', 'end_seq_num']},
                                                                      {'CATEGORY_NAME': 'entity', 'ATTRIBUTE_NAME_LIST': ['rcsb_multiple_source_flag', 'rcsb_source_part_count']},
                                                                      ],
                                 ('GENERATED_CONTENT', 'data_exchange'): [{'CATEGORY_NAME': 'rcsb_data_exchange_status',
                                                                           'ATTRIBUTE_NAME_LIST': ['update_id', 'database_name', 'object_name', 'update_status_flag', 'update_begin_timestamp', 'update_end_timestamp']}],
                                 ('EVOLVING_CONTENT', 'pdbx_core'): [{'CATEGORY_NAME': 'diffrn', 'ATTRIBUTE_NAME_LIST': ['pdbx_serial_crystal_experiment']},
                                                                     {'CATEGORY_NAME': 'diffrn_detector', 'ATTRIBUTE_NAME_LIST': ['pdbx_frequency']},
                                                                     {'CATEGORY_NAME': 'pdbx_serial_crystallography_measurement',
                                                                      'ATTRIBUTE_NAME_LIST': ['diffrn_id',
                                                                                              'pulse_energy',
                                                                                              'pulse_duration',
                                                                                              'xfel_pulse_repetition_rate',
                                                                                              'pulse_photon_energy',
                                                                                              'photons_per_pulse',
                                                                                              'source_size',
                                                                                              'source_distance',
                                                                                              'focal_spot_size',
                                                                                              'collimation',
                                                                                              'collection_time_total']},

                                                                     {'CATEGORY_NAME': 'pdbx_serial_crystallography_sample_delivery',
                                                                      'ATTRIBUTE_NAME_LIST': ['diffrn_id', 'description', 'method']},

                                                                     {'CATEGORY_NAME': 'pdbx_serial_crystallography_sample_delivery_injection',
                                                                      'ATTRIBUTE_NAME_LIST': ['diffrn_id',
                                                                                              'description',
                                                                                              'injector_diameter',
                                                                                              'injector_temperature',
                                                                                              'injector_pressure',
                                                                                              'flow_rate',
                                                                                              'carrier_solvent',
                                                                                              'crystal_concentration',
                                                                                              'preparation',
                                                                                              'power_by',
                                                                                              'injector_nozzle',
                                                                                              'jet_diameter',
                                                                                              'filter_size']},

                                                                     {'CATEGORY_NAME': 'pdbx_serial_crystallography_sample_delivery_fixed_target',
                                                                      'ATTRIBUTE_NAME_LIST': ['diffrn_id',
                                                                                              'description',
                                                                                              'sample_holding',
                                                                                              'support_base',
                                                                                              'sample_unit_size',
                                                                                              'crystals_per_unit',
                                                                                              'sample_solvent',
                                                                                              'sample_dehydration_prevention',
                                                                                              'motion_control',
                                                                                              'velocity_horizontal',
                                                                                              'velocity_vertical',
                                                                                              'details']},

                                                                     {'CATEGORY_NAME': 'pdbx_serial_crystallography_data_reduction',
                                                                      'ATTRIBUTE_NAME_LIST': ['diffrn_id',
                                                                                              'frames_total',
                                                                                              'xfel_pulse_events',
                                                                                              'frame_hits',
                                                                                              'crystal_hits',
                                                                                              'droplet_hits',
                                                                                              'frames_failed_index',
                                                                                              'frames_indexed',
                                                                                              'lattices_indexed',
                                                                                              'xfel_run_numbers']},
                                                                     ],
                                 ('GENERATED_CONTENT', 'chem_comp'): [{'CATEGORY_NAME': 'rcsb_load_status', 'ATTRIBUTE_NAME_LIST': ['datablock_name', 'load_date', 'locator']}],
                                 ('GENERATED_CONTENT', 'bird_chem_comp'): [{'CATEGORY_NAME': 'rcsb_load_status', 'ATTRIBUTE_NAME_LIST': ['datablock_name', 'load_date', 'locator']}],
                                 ('GENERATED_CONTENT', 'bird'): [{'CATEGORY_NAME': 'rcsb_load_status', 'ATTRIBUTE_NAME_LIST': ['datablock_name', 'load_date', 'locator']}],
                                 ('GENERATED_CONTENT', 'bird_family'): [{'CATEGORY_NAME': 'rcsb_load_status', 'ATTRIBUTE_NAME_LIST': ['datablock_name', 'load_date', 'locator']}],
                                 #
                                 ('REPO_HOLDINGS_CONTENT', 'repository_holdings'): [{'CATEGORY_NAME': 'rcsb_repository_holdings_current',
                                                                                     'ATTRIBUTE_NAME_LIST': ['update_id', 'entry_id', 'repository_content_types']},
                                                                                    {'CATEGORY_NAME': 'rcsb_repository_holdings_update',
                                                                                     'ATTRIBUTE_NAME_LIST': ['update_id', 'entry_id', 'repository_content_types']},
                                                                                    {'CATEGORY_NAME': 'rcsb_repository_holdings_removed',
                                                                                     'ATTRIBUTE_NAME_LIST': ['update_id', 'entry_id', 'repository_content_types', 'deposit_date',
                                                                                                             'release_date', 'remove_date', 'title', 'details', 'audit_authors', 'id_codes_replaced_by']},
                                                                                    {'CATEGORY_NAME': 'rcsb_repository_holdings_unreleased',
                                                                                     'ATTRIBUTE_NAME_LIST': ['update_id', 'entry_id', 'status_code', 'deposit_date', 'deposit_date_coordinates',
                                                                                                             'deposit_date_structure_factors', 'deposit_date_nmr_restraints', 'hold_date_coordinates',
                                                                                                             'hold_date_structure_factors', 'hold_date_nmr_restraints', 'release_date', 'title',
                                                                                                             'audit_authors', 'author_prerelease_sequence_status']},
                                                                                    {'CATEGORY_NAME': 'rcsb_repository_holdings_removed_audit_author',
                                                                                     'ATTRIBUTE_NAME_LIST': ['update_id', 'entry_id', 'ordinal_id', 'audit_author']},
                                                                                    {'CATEGORY_NAME': 'rcsb_repository_holdings_superseded',
                                                                                     'ATTRIBUTE_NAME_LIST': ['update_id', 'entry_id', 'id_codes_superseded']},
                                                                                    {'CATEGORY_NAME': 'rcsb_repository_holdings_transferred', 'ATTRIBUTE_NAME_LIST': ['update_id', 'entry_id',
                                                                                                                                                                      'status_code',
                                                                                                                                                                      'deposit_date',
                                                                                                                                                                      'release_date',
                                                                                                                                                                      'title',
                                                                                                                                                                      'audit_authors',
                                                                                                                                                                      'remote_accession_code',
                                                                                                                                                                      'remote_repository_name',
                                                                                                                                                                      'repository_content_types']},
                                                                                    {'CATEGORY_NAME': 'rcsb_repository_holdings_insilico_models', 'ATTRIBUTE_NAME_LIST': ['update_id', 'entry_id',
                                                                                                                                                                          'status_code',
                                                                                                                                                                          'deposit_date',
                                                                                                                                                                          'release_date',
                                                                                                                                                                          'title',
                                                                                                                                                                          'audit_authors',
                                                                                                                                                                          'remove_date',
                                                                                                                                                                          'id_codes_replaced_by']},
                                                                                    ],
                                 ('SEQUENCE_CLUSTER_CONTENT', 'entity_sequence_clusters'): [{'CATEGORY_NAME': 'rcsb_entity_sequence_cluster_list',
                                                                                             'ATTRIBUTE_NAME_LIST': ['data_set_id', 'entry_id', 'entity_id', 'cluster_id', 'identity']},
                                                                                            {'CATEGORY_NAME': 'rcsb_instance_sequence_cluster_list',
                                                                                             'ATTRIBUTE_NAME_LIST': ['data_set_id', 'entry_id', 'instance_id', 'cluster_id', 'identity']}]
                                 }
        #
        #
        # _diffrn.pdbx_serial_crystal_experiment
        # _diffrn_detector.pdbx_frequency
        #
        #
        dH['slice_parent_items'] = {('ENTITY', 'pdbx_core'): [{'CATEGORY_NAME': 'entity', 'ATTRIBUTE_NAME': 'id'}],
                                    ('ASSEMBLY', 'pdbx_core'): [{'CATEGORY_NAME': 'pdbx_struct_assembly', 'ATTRIBUTE_NAME': 'id'}]
                                    }
        dH['slice_parent_filters'] = {('ENTITY', 'pdbx_core'): [{'CATEGORY_NAME': 'entity', 'ATTRIBUTE_NAME': 'type', 'VALUES': ['polymer', 'non-polymer', 'macrolide', 'branched']}]
                                      }

        dH['slice_cardinality_category_extras'] = {('ENTITY', 'pdbx_core'): ['rcsb_load_status', 'rcsb_entity_container_identifiers'],
                                                   ('ASSEMBLY', 'pdbx_core'): ['rcsb_load_status', 'rcsb_assembly_container_identifiers']
                                                   }
        dH['slice_category_extras'] = {('ENTITY', 'pdbx_core'): ['rcsb_load_status'], ('ASSEMBLY', 'pdbx_core'): ['rcsb_load_status', 'pdbx_struct_oper_list']}

        # included or excluded schema identifiers
        #
        sD['schema_content_filters'] = {'pdbx': {'INCLUDE': [],
                                                 'EXCLUDE': ['ATOM_SITE',
                                                             'ATOM_SITE_ANISOTROP'
                                                             ]
                                                 },
                                        'pdbx_core': {'INCLUDE': [],
                                                      'EXCLUDE': ['ATOM_SITE',
                                                                  'ATOM_SITE_ANISOTROP'
                                                                  ]
                                                      },
                                        'repository_holdings': {'INCLUDE': ['rcsb_repository_holdings_update', 'rcsb_repository_holdings_current',
                                                                            'rcsb_repository_holdings_unreleased', 'rcsb_repository_holdings_removed',
                                                                            'rcsb_repository_holdings_removed_audit_author',
                                                                            'rcsb_repository_holdings_superseded',
                                                                            'rcsb_repository_holdings_transferred', 'rcsb_repository_holdings_insilico_models'],
                                                                'EXCLUDE': []
                                                                },
                                        'entity_sequence_clusters': {'INCLUDE': ['rcsb_instance_sequence_cluster_list', 'rcsb_entity_sequence_cluster_list',
                                                                                 'software', 'citation', 'citation_author'],
                                                                     'EXCLUDE': []},
                                        'data_exchange': {'INCLUDE': ['rcsb_data_exchange_status'],
                                                          'EXCLUDE': []},
                                        }

        sD['block_attributes'] = {'pdbx': {'ATTRIBUTE_NAME': 'structure_id', 'CIF_TYPE_CODE': 'code', 'MAX_WIDTH': 12, 'METHOD': 'datablockid()'},
                                  'bird': {'ATTRIBUTE_NAME': 'db_id', 'CIF_TYPE_CODE': 'code', 'MAX_WIDTH': 10, 'METHOD': 'datablockid()'},
                                  'bird_family': {'ATTRIBUTE_NAME': 'db_id', 'CIF_TYPE_CODE': 'code', 'MAX_WIDTH': 10, 'METHOD': 'datablockid()'},
                                  'chem_comp': {'ATTRIBUTE_NAME': 'component_id', 'CIF_TYPE_CODE': 'code', 'MAX_WIDTH': 10, 'METHOD': 'datablockid()'},
                                  'bird_chem_comp': {'ATTRIBUTE_NAME': 'component_id', 'CIF_TYPE_CODE': 'code', 'MAX_WIDTH': 10, 'METHOD': 'datablockid()'},
                                  'pdb_distro': {'ATTRIBUTE_NAME': 'structure_id', 'CIF_TYPE_CODE': 'code', 'MAX_WIDTH': 12, 'METHOD': 'datablockid()'},
                                  }
        sD['database_names'] = {'pdbx': {'NAME': 'pdbx_v5', 'VERSION': '0_2'},
                                'pdbx_core': {'NAME': 'pdbx_v5', 'VERSION': '0_2'},
                                'bird': {'NAME': 'bird_v5', 'VERSION': '0_1'},
                                'bird_family': {'NAME': 'bird_v5', 'VERSION': '0_1'},
                                'chem_comp': {'NAME': 'chem_comp_v5', 'VERSION': '0_1'},
                                'bird_chem_comp': {'NAME': 'chem_comp_v5', 'VERSION': '0_1'},
                                'pdb_distro': {'NAME': 'stat', 'VERSION': '0_1'},
                                'repository_holdings': {'NAME': 'repository_holdings', 'VERSION': 'v5'},
                                'entity_sequence_clusters': {'NAME': 'sequence_clusters', 'VERSION': 'v5'},
                                'data_exchange': {'NAME': 'data_exchange', 'VERSION': 'v5'},
                                }
        #
        #
        dD['schema_collection_names'] = {'pdbx': ['pdbx_v5_0_2', 'pdbx_ext_v5_0_2'],
                                         'pdbx_core': ['pdbx_core_entity_v5_0_2', 'pdbx_core_entry_v5_0_2', 'pdbx_core_assembly_v5_0_2'],
                                         'bird': ['bird_v5_0_2'],
                                         'bird_family': ['family_v5_0_2'],
                                         'chem_comp': ['chem_comp_v5_0_2'],
                                         'bird_chem_comp': ['bird_chem_comp_v5_0_2'],
                                         'pdb_distro': [],
                                         'repository_holdings': ['repository_holdings_update_v0_1', 'repository_holdings_current_v0_1', 'repository_holdings_unreleased_v0_1',
                                                                 'repository_holdings_removed_v0_1', 'repository_holdings_removed_audit_authors_v0_1',
                                                                 'repository_holdings_superseded_v0_1', 'repository_holdings_transferred_v0_1',
                                                                 'rcsb_repository_holdings_insilico_models_v0_1'],
                                         'entity_sequence_clusters': ['cluster_members_v0_1', 'cluster_provenance_v0_1', 'entity_members_v0_1']
                                         }
        #
        # RCSB_LOAD_STATUS must be included in all INCLUDE filters -
        #
        dD['schema_content_filters'] = {'pdbx_v5_0_2': {'INCLUDE': [],
                                                        'EXCLUDE': ['NDB_STRUCT_CONF_NA',
                                                                    'NDB_STRUCT_FEATURE_NA',
                                                                    'NDB_STRUCT_NA_BASE_PAIR',
                                                                    'NDB_STRUCT_NA_BASE_PAIR_STEP',
                                                                    'PDBX_VALIDATE_CHIRAL',
                                                                    'PDBX_VALIDATE_CLOSE_CONTACT',
                                                                    'PDBX_VALIDATE_MAIN_CHAIN_PLANE',
                                                                    'PDBX_VALIDATE_PEPTIDE_OMEGA',
                                                                    'PDBX_VALIDATE_PLANES',
                                                                    'PDBX_VALIDATE_PLANES_ATOM',
                                                                    'PDBX_VALIDATE_POLYMER_LINKAGE',
                                                                    'PDBX_VALIDATE_RMSD_ANGLE',
                                                                    'PDBX_VALIDATE_RMSD_BOND',
                                                                    'PDBX_VALIDATE_SYMM_CONTACT',
                                                                    'PDBX_VALIDATE_TORSION',
                                                                    'STRUCT_SHEET',
                                                                    'STRUCT_SHEET_HBOND',
                                                                    'STRUCT_SHEET_ORDER',
                                                                    'STRUCT_SHEET_RANGE',
                                                                    'STRUCT_CONF',
                                                                    'STRUCT_CONF_TYPE',
                                                                    'STRUCT_CONN',
                                                                    'STRUCT_CONN_TYPE',
                                                                    'ATOM_SITE',
                                                                    'ATOM_SITE_ANISOTROP',
                                                                    'PDBX_UNOBS_OR_ZERO_OCC_ATOMS',
                                                                    'PDBX_UNOBS_OR_ZERO_OCC_RESIDUES'],
                                                        'SLICE': None
                                                        },

                                        'pdbx_ext_v5_0_2': {'INCLUDE': ['ENTRY', 'NDB_STRUCT_CONF_NA',
                                                                        'NDB_STRUCT_FEATURE_NA',
                                                                        'NDB_STRUCT_NA_BASE_PAIR',
                                                                        'NDB_STRUCT_NA_BASE_PAIR_STEP',
                                                                        'PDBX_VALIDATE_CHIRAL',
                                                                        'PDBX_VALIDATE_CLOSE_CONTACT',
                                                                        'PDBX_VALIDATE_MAIN_CHAIN_PLANE',
                                                                        'PDBX_VALIDATE_PEPTIDE_OMEGA',
                                                                        'PDBX_VALIDATE_PLANES',
                                                                        'PDBX_VALIDATE_PLANES_ATOM',
                                                                        'PDBX_VALIDATE_POLYMER_LINKAGE',
                                                                        'PDBX_VALIDATE_RMSD_ANGLE',
                                                                        'PDBX_VALIDATE_RMSD_BOND',
                                                                        'PDBX_VALIDATE_SYMM_CONTACT',
                                                                        'PDBX_VALIDATE_TORSION',
                                                                        'STRUCT_SHEET',
                                                                        'STRUCT_SHEET_HBOND',
                                                                        'STRUCT_SHEET_ORDER',
                                                                        'STRUCT_SHEET_RANGE',
                                                                        'STRUCT_CONF',
                                                                        'STRUCT_CONF_TYPE',
                                                                        'STRUCT_CONN',
                                                                        'STRUCT_CONN_TYPE',
                                                                        'RCSB_LOAD_STATUS'],
                                                            'EXCLUDE': [],
                                                            'SLICE': None
                                                            },
                                        'pdbx_core_entity_v5_0_2': {'INCLUDE': [], 'EXCLUDE': [], 'SLICE': 'ENTITY'},
                                        'pdbx_core_assembly_v5_0_2': {'INCLUDE': [], 'EXCLUDE': [], 'SLICE': 'ASSEMBLY'},
                                        'pdbx_core_entry_v5_0_2': {'INCLUDE': ['AUDIT_AUTHOR', 'CELL',
                                                                               'CITATION', 'CITATION_AUTHOR', 'DIFFRN', 'DIFFRN_DETECTOR', 'DIFFRN_RADIATION', 'DIFFRN_SOURCE', 'EM_2D_CRYSTAL_ENTITY',
                                                                               'EM_3D_CRYSTAL_ENTITY', 'EM_3D_FITTING', 'EM_3D_RECONSTRUCTION', 'EM_EMBEDDING', 'EM_ENTITY_ASSEMBLY', 'EM_EXPERIMENT',
                                                                               'EM_HELICAL_ENTITY', 'EM_IMAGE_RECORDING', 'EM_IMAGING', 'EM_SINGLE_PARTICLE_ENTITY', 'EM_SOFTWARE', 'EM_SPECIMEN',
                                                                               'EM_STAINING', 'EM_VITRIFICATION', 'ENTRY', 'EXPTL', 'EXPTL_CRYSTAL_GROW', 'PDBX_AUDIT_REVISION_DETAILS',
                                                                               'PDBX_AUDIT_REVISION_HISTORY', 'PDBX_AUDIT_SUPPORT', 'PDBX_DATABASE_PDB_OBS_SPR', 'PDBX_DATABASE_STATUS', 'PDBX_DEPOSIT_GROUP',
                                                                               'PDBX_MOLECULE', 'PDBX_MOLECULE_FEATURES', 'PDBX_NMR_DETAILS', 'PDBX_NMR_ENSEMBLE', 'PDBX_NMR_EXPTL', 'PDBX_NMR_EXPTL_SAMPLE_CONDITIONS',
                                                                               'PDBX_NMR_REFINE', 'PDBX_NMR_SAMPLE_DETAILS', 'PDBX_NMR_SOFTWARE', 'PDBX_NMR_SPECTROMETER', 'PDBX_SG_PROJECT',
                                                                               'RCSB_ATOM_COUNT', 'RCSB_BINDING', 'RCSB_EXTERNAL_REFERENCES', 'RCSB_HAS_CHEMICAL_SHIFT_FILE', 'RCSB_HAS_ED_MAP_FILE',
                                                                               'RCSB_HAS_FOFC_FILE', 'RCSB_HAS_NMR_V1_FILE', 'RCSB_HAS_NMR_V2_FILE', 'RCSB_HAS_STRUCTURE_FACTORS_FILE', 'RCSB_HAS_TWOFOFC_FILE',
                                                                               'RCSB_HAS_VALIDATION_REPORT', 'RCSB_LATEST_REVISION', 'RCSB_MODELS_COUNT', 'RCSB_MOLECULAR_WEIGHT', 'RCSB_PUBMED', 'RCSB_RELEASE_DATE',
                                                                               'REFINE', 'REFINE_ANALYZE', 'REFINE_HIST', 'REFINE_LS_RESTR', 'REFLNS', 'REFLNS_SHELL', 'SOFTWARE', 'STRUCT', 'STRUCT_KEYWORDS', 'SYMMETRY',
                                                                               'RCSB_LOAD_STATUS', 'RCSB_ENTRY_CONTAINER_IDENTIFIERS', 'PDBX_SERIAL_CRYSTALLOGRAPHY_MEASUREMENT',
                                                                               'PDBX_SERIAL_CRYSTALLOGRAPHY_SAMPLE_DELIVERY',
                                                                               'PDBX_SERIAL_CRYSTALLOGRAPHY_SAMPLE_DELIVERY_INJECTION',
                                                                               'PDBX_SERIAL_CRYSTALLOGRAPHY_SAMPLE_DELIVERY_FIXED_TARGET',
                                                                               'PDBX_SERIAL_CRYSTALLOGRAPHY_DATA_REDUCTION'],
                                                                   'EXCLUDE': [], 'SLICE': None},
                                        'repository_holdings_update_v0_1': {'INCLUDE': ['rcsb_repository_holdings_update'], 'EXCLUDE': [], 'SLICE': None},
                                        'repository_holdings_current_v0_1': {'INCLUDE': ['rcsb_repository_holdings_current'], 'EXCLUDE': [], 'SLICE': None},
                                        'repository_holdings_unreleased_v0_1': {'INCLUDE': ['rcsb_repository_holdings_unreleased'], 'EXCLUDE': [], 'SLICE': None},
                                        'repository_holdings_removed_v0_1': {'INCLUDE': ['rcsb_repository_holdings_removed'], 'EXCLUDE': [], 'SLICE': None},
                                        'repository_holdings_removed_audit_authors_v0_1': {'INCLUDE': ['rcsb_repository_holdings_removed_audit_author'], 'EXCLUDE': [], 'SLICE': None},
                                        'repository_holdings_superseded_v0_1': {'INCLUDE': ['rcsb_repository_holdings_superseded'], 'EXCLUDE': [], 'SLICE': None},
                                        'repository_holdings_transferred_v0_1': {'INCLUDE': ['rcsb_repository_holdings_transferred'], 'EXCLUDE': [], 'SLICE': None},
                                        'repository_holdings_insilico_models_v0_1': {'INCLUDE': ['rcsb_repository_holdings_insilico_models'], 'EXCLUDE': [], 'SLICE': None},
                                        'cluster_members_v0_1': {'INCLUDE': ['rcsb_entity_sequence_cluster_list'], 'EXCLUDE': [], 'SLICE': None},
                                        'cluster_provenance_v0_1': {'INCLUDE': ['software', 'citation', 'citation_author'], 'EXCLUDE': [], 'SLICE': None},
                                        'entity_members_v0_1': {'INCLUDE': ['rcsb_entity_sequence_cluster_list'], 'EXCLUDE': [], 'SLICE': None},
                                        'rcsb_data_exchange_status_v0_1': {'INCLUDE': ['rcsb_data_exchange_status'], 'EXCLUDE': [], 'SLICE': None},
                                        }

        dD['collection_attribute_names'] = {'pdbx_v5_0_2': ['entry.id'],
                                            'pdbx_ext_v5_0_2': ['entry.id'],
                                            'pdbx_core_entity_v5_0_2': ['entry.id', 'entity.id'],
                                            'pdbx_core_assembly_v5_0_2': ['entry.id', 'pdbx_struct_assembly.id'],
                                            'pdbx_core_entry_v5_0_2': ['entry.id'],
                                            'bird_v5_0_2': ['pdbx_reference_molecule.prd_id'],
                                            'family_v5_0_2': ['pdbx_reference_molecule_family.family_prd_id'],
                                            'chem_comp_v5_0_2': ['chem_comp.component_id'],
                                            'bird_chem_comp_v5_0_2': ['chem_comp.component_id'],
                                            'repository_holdings_update_v0_1': ['update_id'],
                                            'repository_holdings_current_v0_1': ['update_id'],
                                            'repository_holdings_unreleased_v0_1': ['update_id'],
                                            'repository_holdings_removed_v0_1': ['update_id'],
                                            'repository_holdings_removed_audit_authors': ['update_id'],
                                            'repository_holdings_superseded_v0_1': ['update_id'],
                                            'repository_holdings_transferred_v0_1': ['update_id'],
                                            'repository_holdings_insilico_models_v0_1': ['update_id'],
                                            'cluster_members_v0_1': ['update_id'],
                                            'cluster_provenance_v0_1': ['software.name'],
                                            'entity_members_v0_1': ['update_id'],
                                            'rcsb_data_exchange_status_v0_1': ['update_id', 'database_name', 'object_name'],
                                            }
        #
        dD['collection_subcategory_aggregates'] = {'cluster_members_v0_1': ['sequence_membership'],
                                                   'entity_members_v0_1': ['cluster_membership']
                                                   }
        #
        return cfgD


if __name__ == '__main__':
    mc = MakeConfig()
    #



#
# File:    SchemaDefReShape.py
# Author:  J. Westbrook
# Date:    4-Aug-2018
#
#
# Updates:
#    15-Aug-2018 jdw add next() method for py2 compatibility
#    16-Aug-2018 jdw add remaining layouts for sliced schema
##
"""
Companion class to reshape data objects produced by SchemaDefDataPrep()

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import itertools
import logging

logger = logging.getLogger(__name__)


class SliceValues(object):
    """Iterator class for parent slice values.
    """

    def __init__(self, schemaDataDictById, schemaDefObj, sliceFilter):
        self.__sD = schemaDefObj
        spiL = self.__sD.getSliceParentItems(sliceFilter)
        spfL = self.__sD.getSliceParentFilters(sliceFilter)
        vD = {}
        for sp in spiL:
            catId = sp['CATEGORY']
            atId = sp['ATTRIBUTE']
            vals = []
            if catId in schemaDataDictById:
                for rowD in schemaDataDictById[catId]:
                    if self.__testFilter(rowD, catId, spfL):
                        vals.append(((catId, atId), rowD[atId]))
            vD[(catId, atId)] = vals
        #
        self.index = 0
        logger.debug("Parent value dict %r" % vD.items())
        # Make a list of lists and then get the product
        #
        valueL = []
        for ky, vL in vD.items():
            valueL.append(vL)
        self.data = list(itertools.product(*valueL))
        logger.debug("Parent value product list %r" % self.data)

    def __testFilter(self, rowD, catId, filters):
        ok = True
        for filter in filters:
            if catId != filter['CATEGORY']:
                ok = False
                break
            if filter['ATTRIBUTE'] in rowD and rowD[filter['ATTRIBUTE']] in filter['VALUES']:
                continue
            else:
                ok = False
                break

        return ok

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        try:
            result = self.data[self.index]
            self.index += 1
        except IndexError:
            raise StopIteration
        return result

    def __prev__(self):
        self.index -= 1
        if self.index < 0:
            raise StopIteration
        return self.data[self.index]


class SchemaDefReShape(object):

    """Companion class to reshape data objects produced by SchemaDefDataPrep()
    """

    def __init__(self, schemaDefAccessObj, workPath=None, verbose=True):
        self.__verbose = verbose
        self.__workPath = workPath
        self.__debug = False
        self.__sD = schemaDefAccessObj
        #

    def applyShape(self, schemaDataDictById, styleType="rowwise_by_name"):
        """
        """
        return self.__reshapeSchemaData(schemaDataDictById, styleType=styleType)

    def applySlicedShape(self, schemaDataDictById, styleType="rowwise_by_name", logSize=False, sliceFilter=None):
        """
        """
        rL = []
        if sliceFilter:
            rL = []
            sI = self.__sD.getSliceIndex(sliceFilter)
            logger.debug("Slice index %r" % sI.items())
            #
            sliceValues = SliceValues(schemaDataDictById, self.__sD, sliceFilter)
            for ii, sliceValue in enumerate(sliceValues):
                logger.debug(" %4d filter %s slice value %r" % (ii, sliceFilter, sliceValue))
                rD = self.__reshapeSlidedSchemaData(schemaDataDictById, sliceFilter, sliceValue, sI, styleType=styleType)
                rL.append(rD)
        else:
            return [self.__reshapeSchemaData(schemaDataDictById, styleType=styleType)]

        return rL

    def __reshapeSlidedSchemaData(self, schemaDataDictById, sliceFilter, sliceValues, sliceIndex, styleType="rowwise_by_name"):
        """  Reorganize and rename input table data object according to the input style preference:

             Input: schemaDataDictById  (styleType="rowwise_by_id")
                                         dict[<schemaId>]   = [ row1asDict[attributeId]=value,  row2asDict[attribute]=value, .. ]

             Output: rowwise_by_name:     dict[<schemaObjName>] = [ row1Dict[attributeName]=value,  row2dict[], .. ]
                     rowwise_no_name:    dict[<schemaObjName>] = {'attributes': [atName1, atName2,... ], 'data' : [[val1, val2, .. ],[val1, val2,... ]]}
                     columnwise_by_name: dict[<schemaObjName>] = {'atName': [val1, val2,... ], atName2: [val1,val2, ... ], ...}
      rowwise_by_name_with_cardinality:  same as rowwise_byName with special handing of tables with unit cardinality
                                                          dict[<schemaObjName>] = row1Dict[attributeName]=value (singleton row)
        """
        rD = {}
        try:
            if styleType == "rowwise_by_name":
                rD = self.__sliceRowwiseByName(schemaDataDictById, sliceFilter, sliceValues, sliceIndex)
            elif styleType == "rowwise_by_name_with_cardinality":
                rD = self.__sliceRowwiseByNameWithCard(schemaDataDictById, sliceFilter, sliceValues, sliceIndex)
            elif styleType == "columnwise_by_name":
                rD = self.__sliceColumnwise(schemaDataDictById, sliceFilter, sliceValues, sliceIndex)
            elif styleType == "rowwise_no_name":
                rD = self.__shapeRowwiseNoName(schemaDataDictById)
            elif styleType == "rowwise_by_id":
                rD = schemaDataDictById
            else:
                rD = schemaDataDictById
                logger.warning("Unsupported style type %s" % styleType)
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
            rD = schemaDataDictById

        return rD

    def __inSlice(self, schemaId, sliceIdx, rowD, parentVals):
        """ Test if the child values in the input row dictionary equal the corresponding input parents values
        """
        ok = True
        try:
            if schemaId in sliceIdx:
                for parentVal in parentVals:
                    if parentVal[0] in sliceIdx[schemaId]:
                        cAtId = sliceIdx[schemaId][parentVal[0]]
                        if cAtId in rowD and rowD[cAtId] == parentVal[1]:
                            continue
                        else:
                            ok = False
                            break
                    else:
                        ok = False
                        break
            else:
                ok = False

        except Exception as e:
            logger.exception("Failing with %s" % str(e))
            ok = False
        return ok

    def __sliceRowwiseByNameWithCard(self, schemaDataDictById, sliceFilter, sliceValues, sliceIndex):
        rD = {}
        for schemaId in schemaDataDictById:
            schemaObj = self.__sD.getSchemaObject(schemaId)
            lExtra = schemaObj.isSliceExtra(sliceFilter)
            if lExtra:
                logger.debug("SchemaId %r slice extra %r" % (schemaId, lExtra))
            if schemaId not in sliceIndex and not schemaObj.isSliceExtra(sliceFilter):
                continue
            # ------
            schemaObjName = self.__sD.getSchemaName(schemaId)
            #
            iRowDList = schemaDataDictById[schemaId]

            #
            oRowDList = []
            for iRowD in iRowDList:
                if lExtra or self.__inSlice(schemaId, sliceIndex, iRowD, sliceValues):
                    oRowD = {}
                    for atId in iRowD:
                        oRowD[schemaObj.getAttributeName(atId)] = iRowD[atId]
                    oRowDList.append(oRowD)

            if (len(oRowDList) < 1) and not schemaObj.isMandatory():
                logger.debug("Schema id %r row length %r mandatory %r" % (schemaId, len(oRowDList), schemaObj.isMandatory()))
                continue

            if schemaObj.hasSliceUnitCardinality(sliceFilter) and len(oRowDList) == 1:
                rD[schemaObjName] = oRowDList[0]
            else:
                rD[schemaObjName] = oRowDList

        return rD

    def __sliceRowwiseByName(self, schemaDataDictById, sliceFilter, sliceValues, sliceIndex):
        rD = {}
        for schemaId in schemaDataDictById:
            schemaObj = self.__sD.getSchemaObject(schemaId)
            lExtra = schemaObj.isSliceExtra(sliceFilter)
            if lExtra:
                logger.debug("SchemaId %r slice extra %r" % (schemaId, lExtra))
            if schemaId not in sliceIndex and not schemaObj.isSliceExtra(sliceFilter):
                continue
            # ------
            schemaObjName = self.__sD.getSchemaName(schemaId)
            #
            iRowDList = schemaDataDictById[schemaId]

            oRowDList = []
            for iRowD in iRowDList:
                if lExtra or self.__inSlice(schemaId, sliceIndex, iRowD, sliceValues):
                    oRowD = {}
                    for atId in iRowD:
                        oRowD[schemaObj.getAttributeName(atId)] = iRowD[atId]
                    oRowDList.append(oRowD)
            #
            if (len(oRowDList) < 1) and not schemaObj.isMandatory():
                logger.debug("Schema id %r row length %r mandatory %r" % (schemaId, len(oRowDList), schemaObj.isMandatory()))
                continue
            rD[schemaObjName] = oRowDList

        return rD

    def __sliceColumnwise(self, schemaDataDictById, sliceFilter, sliceValues, sliceIndex):
        rD = {}
        for schemaId in schemaDataDictById:
            schemaObj = self.__sD.getSchemaObject(schemaId)
            lExtra = schemaObj.isSliceExtra(sliceFilter)
            if lExtra:
                logger.debug("SchemaId %r slice extra %r" % (schemaId, lExtra))
            if schemaId not in sliceIndex and not schemaObj.isSliceExtra(sliceFilter):
                continue
            # ------
            schemaObjName = self.__sD.getSchemaName(schemaId)
            iRowDList = schemaDataDictById[schemaId]
            #
            colD = {}
            for iRowD in iRowDList:
                if lExtra or self.__inSlice(schemaId, sliceIndex, iRowD, sliceValues):
                    for atId in iRowD:
                        atName = schemaObj.getAttributeName(atId)
                        if atName not in colD:
                            colD[atName] = []
                        colD[atName].append(iRowD[atId])

            if (len(colD) < 1) and not schemaObj.isMandatory():
                logger.debug("Schema id %r row length %r mandatory %r" % (schemaId, len(colD), schemaObj.isMandatory()))
                continue
            rD[schemaObjName] = colD

        return rD

    def __sliceRowwiseNoName(self, schemaDataDictById, sliceFilter, sliceValues, sliceIndex):
        rD = {}
        for schemaId in schemaDataDictById:
            schemaObj = self.__sD.getSchemaObject(schemaId)
            lExtra = schemaObj.isSliceExtra(sliceFilter)
            if lExtra:
                logger.debug("SchemaId %r slice extra %r" % (schemaId, lExtra))
            if schemaId not in sliceIndex and not schemaObj.isSliceExtra(sliceFilter):
                continue
            # ------
            atIdList = self.__sD.getAttributeIdList(schemaId)
            atNameList = self.__sD.getAttributeNameList(schemaId)
            schemaObjName = self.__sD.getSchemaName(schemaId)
            #
            iRowDList = schemaDataDictById[schemaId]

            #
            oRowList = []
            for iRowD in iRowDList:
                if lExtra or self.__inSlice(schemaId, sliceIndex, iRowD, sliceValues):
                    oRowL = []
                    for atId in atIdList:
                        oRowL.append(iRowD[atId])
                    oRowList.append(oRowL)
            #
            if (len(oRowList) < 1) and not schemaObj.isMandatory():
                logger.debug("Schema id %r row length %r mandatory %r" % (schemaId, len(oRowList), schemaObj.isMandatory()))
                continue
            rD[schemaObjName] = {'attributes': atNameList, 'data': oRowList}

        return rD

    def __reshapeSchemaData(self, schemaDataDictById, styleType="rowwise_by_name"):
        """  Reorganize and rename input table data object according to the input style preference:

             Input: schemaDataDictById  (styleType="rowwise_by_id")
                                         dict[<schemaId>]   = [ row1asDict[attributeId]=value,  row2asDict[attribute]=value, .. ]

             Output: rowwise_by_name:    dict[<schemaObjName>] = [ row1Dict[attributeName]=value,  row2dict[], .. ]
                     rowwise_no_name:    dict[<schemaObjName>] = {'attributes': [atName1, atName2,... ], 'data' : [[val1, val2, .. ],[val1, val2,... ]]}
                     columnwise_by_name: dict[<schemaObjName>] = {'atName': [val1, val2,... ], atName2: [val1,val2, ... ], ...}
      rowwise_by_name_with_cardinality:  same as rowwise_byName with special handing of tables with unit cardinality
                                                          dict[<schemaObjName>] = row1Dict[attributeName]=value (singleton row)
        """
        rD = {}
        try:
            if styleType == "rowwise_by_name":
                rD = self.__shapeRowwiseByName(schemaDataDictById)
            elif styleType == "rowwise_by_name_with_cardinality":
                rD = self.__shapeRowwiseByNameWithCard(schemaDataDictById)
            elif styleType == "columnwise_by_name":
                rD = self.__shapeColumnwise(schemaDataDictById)
            elif styleType == "rowwise_no_name":
                rD = self.__shapeRowwiseNoName(schemaDataDictById)
            elif styleType == "rowwise_by_id":
                rD = schemaDataDictById
            else:
                rD = schemaDataDictById
                logger.warning("Unsupported style type %s" % styleType)
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
            rD = schemaDataDictById

        return rD

    def __shapeRowwiseByName(self, schemaDataDictById):
        rD = {}
        for schemaId in schemaDataDictById:
            schemaObj = self.__sD.getSchemaObject(schemaId)
            schemaObjName = self.__sD.getSchemaName(schemaId)
            iRowDList = schemaDataDictById[schemaId]
            oRowDList = []
            for iRowD in iRowDList:
                oRowD = {}
                for atId in iRowD:
                    oRowD[schemaObj.getAttributeName(atId)] = iRowD[atId]
                oRowDList.append(oRowD)
            rD[schemaObjName] = oRowDList
        return rD

    def __shapeRowwiseByNameWithCard(self, schemaDataDictById):
        rD = {}
        for schemaId in schemaDataDictById:
            schemaObj = self.__sD.getSchemaObject(schemaId)
            schemaObjName = self.__sD.getSchemaName(schemaId)
            unitCard = self.__sD.hasUnitCardinality(schemaId)
            iRowDList = schemaDataDictById[schemaId]
            #
            if unitCard and len(iRowDList) == 1:
                iRowD = iRowDList[0]
                oRowD = {}
                for atId in iRowD:
                    oRowD[schemaObj.getAttributeName(atId)] = iRowD[atId]
                rD[schemaObjName] = oRowD
            else:
                oRowDList = []
                for iRowD in iRowDList:
                    oRowD = {}
                    for atId in iRowD:
                        oRowD[schemaObj.getAttributeName(atId)] = iRowD[atId]
                    oRowDList.append(oRowD)
                rD[schemaObjName] = oRowDList
        return rD

    def __shapeColumnwise(self, schemaDataDictById):
        rD = {}
        for schemaId in schemaDataDictById:
            schemaObj = self.__sD.getSchemaObject(schemaId)
            schemaObjName = self.__sD.getSchemaName(schemaId)
            iRowDList = schemaDataDictById[schemaId]
            colD = {}
            for iRowD in iRowDList:
                for atId in iRowD:
                    atName = schemaObj.getAttributeName(atId)
                    if atName not in colD:
                        colD[atName] = []
                    colD[atName].append(iRowD[atId])
            rD[schemaObjName] = colD
        return rD

    def __shapeRowwiseNoName(self, schemaDataDictById):
        rD = {}
        for schemaId in schemaDataDictById:
            # schemaObj = self.__sD.getSchemaObject(schemaId)
            schemaObjName = self.__sD.getSchemaName(schemaId)
            atIdList = self.__sD.getAttributeIdList(schemaId)
            atNameList = self.__sD.getAttributeNameList(schemaId)
            #
            iRowDList = schemaDataDictById[schemaId]
            oRowList = []
            for iRowD in iRowDList:
                oRowL = []
                for atId in atIdList:
                    oRowL.append(iRowD[atId])
                oRowList.append(oRowL)
            #
            rD[schemaObjName] = {'attributes': atNameList, 'data': oRowList}
        return rD



##
# File: RepoHoldingsEtlWorker.py
# Date: 2-Jul-2018  jdw
#
# ETL utilities for processing repository holding data.
# Updates:
#  15-Jul-2018 jdw split out to separate module and add status tracking
#  26-Nov-2018 jdw add COLLECTION_HOLDINGS_PRERELEASE
#
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

from rcsb.db.mongo.DocumentLoader import DocumentLoader
from rcsb.db.processors.DataExchangeStatus import DataExchangeStatus
from rcsb.db.processors.RepoHoldingsDataPrep import RepoHoldingsDataPrep

logger = logging.getLogger(__name__)


class RepoHoldingsEtlWorker(object):
    """ Prepare and load repository holdings and repository update data.
    """

    def __init__(self, cfgOb, sandboxPath, workPath=None, numProc=2, chunkSize=10, readBackCheck=False, documentLimit=None, verbose=False):
        self.__cfgOb = cfgOb
        self.__sandboxPath = sandboxPath
        self.__workPath = workPath
        self.__readBackCheck = readBackCheck
        self.__numProc = numProc
        self.__chunkSize = chunkSize
        self.__documentLimit = documentLimit
        self.__resourceName = "MONGO_DB"
        self.__filterType = "assign-dates"
        self.__verbose = verbose
        self.__statusList = []

    def __updateStatus(self, updateId, databaseName, collectionName, status, startTimestamp):
        try:
            sFlag = 'Y' if status else 'N'
            desp = DataExchangeStatus()
            desp.setStartTime(tS=startTimestamp)
            desp.setObject(databaseName, collectionName)
            desp.setStatus(updateId=updateId, successFlag=sFlag)
            desp.setEndTime()
            self.__statusList.append(desp.getStatus())
            return True
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return False

    def load(self, updateId, loadType='full'):
        """ Test case - load legacy repository holdings and status data -

        Relevant configuration options:

        [DEFAULT]
        RCSB_EXCHANGE_SANDBOX_PATH=MOCK_EXCHANGE_SANDBOX

        [repository_holdings]
        DATABASE_NAME=repository_holdings
        DATABASE_VERSION_STRING=v5
        COLLECTION_HOLDINGS_UPDATE=rcsb_repository_holdings_update
        COLLECTION_HOLDINGS_CURRENT=rcsb_repository_holdings_current
        COLLECTION_HOLDINGS_UNRELEASED=rcsb_repository_holdings_unreleased
        COLLECTION_HOLDINGS_PRERELEASE=rcsb_repository_holdings_prerelease
        COLLECTION_HOLDINGS_REMOVED=rcsb_repository_holdings_removed
        COLLECTION_HOLDINGS_REMOVED_AUTHORS=rcsb_repository_holdings_removed_audit_authors
        COLLECTION_HOLDINGS_SUPERSEDED=rcsb_repository_holdings_superseded
        COLLECTION_VERSION_STRING=v0_1

        """
        try:
            self.__statusList = []
            desp = DataExchangeStatus()
            statusStartTimestamp = desp.setStartTime()

            sectionName = 'repository_holdings'

            rhdp = RepoHoldingsDataPrep(sandboxPath=self.__sandboxPath, workPath=self.__workPath, filterType=self.__filterType)
            #
            dl = DocumentLoader(self.__cfgOb, self.__resourceName, numProc=self.__numProc, chunkSize=self.__chunkSize,
                                documentLimit=self.__documentLimit, verbose=self.__verbose, readBackCheck=self.__readBackCheck)
            #
            databaseName = self.__cfgOb.get('DATABASE_NAME', sectionName=sectionName) + '_' + self.__cfgOb.get('DATABASE_VERSION_STRING', sectionName=sectionName)
            collectionVersion = self.__cfgOb.get('COLLECTION_VERSION_STRING', sectionName=sectionName)
            #
            dList = rhdp.getHoldingsUpdate(updateId=updateId)
            collectionName = self.__cfgOb.get('COLLECTION_HOLDINGS_UPDATE', sectionName=sectionName) + '_' + collectionVersion
            ok = dl.load(databaseName, collectionName, loadType=loadType, documentList=dList,
                         indexAttributeList=['update_id', 'entry_id'], keyNames=None)
            self.__updateStatus(updateId, databaseName, collectionName, ok, statusStartTimestamp)
            #
            dList = rhdp.getHoldingsCurrent(updateId=updateId)
            collectionName = self.__cfgOb.get('COLLECTION_HOLDINGS_CURRENT', sectionName=sectionName) + '_' + collectionVersion
            ok = dl.load(databaseName, collectionName, loadType=loadType, documentList=dList,
                         indexAttributeList=['update_id', 'entry_id'], keyNames=None)
            self.__updateStatus(updateId, databaseName, collectionName, ok, statusStartTimestamp)

            dList = rhdp.getHoldingsUnreleased(updateId=updateId)
            collectionName = self.__cfgOb.get('COLLECTION_HOLDINGS_UNRELEASED', sectionName=sectionName) + '_' + collectionVersion
            ok = dl.load(databaseName, collectionName, loadType=loadType, documentList=dList,
                         indexAttributeList=['update_id', 'entry_id'], keyNames=None)
            self.__updateStatus(updateId, databaseName, collectionName, ok, statusStartTimestamp)
            #
            dList = rhdp.getHoldingsPrerelease(updateId=updateId)
            collectionName = self.__cfgOb.get('COLLECTION_HOLDINGS_PRERELEASE', sectionName=sectionName) + '_' + collectionVersion
            ok = dl.load(databaseName, collectionName, loadType=loadType, documentList=dList,
                         indexAttributeList=['update_id', 'entry_id'], keyNames=None)
            self.__updateStatus(updateId, databaseName, collectionName, ok, statusStartTimestamp)
            #
            dList1, dList2, dList3 = rhdp.getHoldingsRemoved(updateId=updateId)
            collectionName = self.__cfgOb.get('COLLECTION_HOLDINGS_REMOVED', sectionName=sectionName) + '_' + collectionVersion
            ok = dl.load(databaseName, collectionName, loadType=loadType, documentList=dList1,
                         indexAttributeList=['update_id', 'entry_id'], keyNames=None)
            self.__updateStatus(updateId, databaseName, collectionName, ok, statusStartTimestamp)

            collectionName = self.__cfgOb.get('COLLECTION_HOLDINGS_REMOVED_AUTHORS', sectionName=sectionName) + '_' + collectionVersion
            ok = dl.load(databaseName, collectionName, loadType=loadType, documentList=dList2,
                         indexAttributeList=['update_id', 'entry_id'], keyNames=None)
            self.__updateStatus(updateId, databaseName, collectionName, ok, statusStartTimestamp)

            collectionName = self.__cfgOb.get('COLLECTION_HOLDINGS_SUPERSEDED', sectionName=sectionName) + '_' + collectionVersion
            ok = dl.load(databaseName, collectionName, loadType=loadType, documentList=dList3,
                         indexAttributeList=['update_id', 'entry_id'], keyNames=None)
            self.__updateStatus(updateId, databaseName, collectionName, ok, statusStartTimestamp)

#
            dList1, dList2 = rhdp.getHoldingsTransferred(updateId=updateId)

            collectionName = self.__cfgOb.get('COLLECTION_HOLDINGS_TRANSFERRED', sectionName=sectionName) + '_' + collectionVersion
            ok = dl.load(databaseName, collectionName, loadType=loadType, documentList=dList1,
                         indexAttributeList=['update_id', 'entry_id'], keyNames=None)
            self.__updateStatus(updateId, databaseName, collectionName, ok, statusStartTimestamp)

            collectionName = self.__cfgOb.get('COLLECTION_HOLDINGS_INSILICO_MODELS', sectionName=sectionName) + '_' + collectionVersion
            ok = dl.load(databaseName, collectionName, loadType=loadType, documentList=dList2,
                         indexAttributeList=['update_id', 'entry_id'], keyNames=None)
            self.__updateStatus(updateId, databaseName, collectionName, ok, statusStartTimestamp)

            return True
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return False

    def getLoadStatus(self):
        return self.__statusList



##
# File:    SchemaDocumentHelper.py
# Author:  J. Westbrook
# Date:    7-Jun-2018
# Version: 0.001 Initial version
#
# Updates:
#  22-Jun-2018 jdw change collection attribute id specification to dot notation
#  14-Aug-2018 jdw generalize document key attribute to attribute list
#  20-Aug-2018 jdw slice details added to __schemaContentFilters
#   8-Oct-2018 jdw added getSubCategoryAggregates() method
#
##
"""
Inject additional document information into a schema definition.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

from rcsb.db.helpers.SchemaDocumentHelperBase import SchemaDocumentHelperBase

logger = logging.getLogger(__name__)


class SchemaDocumentHelper(SchemaDocumentHelperBase):
    """ Inject additional document information into a schema definition.

        Single source of additional document schema semantic content.
    """

    def __init__(self, **kwargs):
        """
        Args:
            **kwargs: (below)

        """
        super(SchemaDocumentHelper, self).__init__(**kwargs)
        # ----
        #
        self.__cfgOb = kwargs.get('cfgOb', None)
        sectionName = kwargs.get('config_section', 'document_helper')
        self.__cfgD = self.__cfgOb.exportConfig(sectionName=sectionName)
        #
        # ----

    def getCollections(self, schemaName):
        cL = []
        try:
            cL = self.__cfgD['schema_collection_names'][schemaName]
        except Exception as e:
            logger.debug("Schema definitiona name %s failing with %s" % (schemaName, str(e)))
        return cL

    def getCollectionMap(self):
        try:
            return self.__cfgD['schema_collection_names']
        except Exception as e:
            logger.debug("Failing with %s" % (str(e)))
        return {}

    def getExcluded(self, collectionName):
        '''  For input collection, return the list of excluded schema identifiers.

        '''
        includeL = []
        try:
            includeL = [tS.upper() for tS in self.__cfgD['schema_content_filters'][collectionName]['EXCLUDE']]
        except Exception as e:
            logger.debug("Collection %s failing with %s" % (collectionName, str(e)))
        return includeL

    def getIncluded(self, collectionName):
        '''  For input collection, return the list of included schema identifiers.

        '''
        excludeL = []
        try:
            excludeL = [tS.upper() for tS in self.__cfgD['schema_content_filters'][collectionName]['INCLUDE']]
        except Exception as e:
            logger.debug("Collection %s failing with %s" % (collectionName, str(e)))
        return excludeL

    def getSliceFilter(self, collectionName):
        '''  For input collection, return an optional slice filter or None.

        '''
        sf = None
        try:
            sf = self.__cfgD['schema_content_filters'][collectionName]['SLICE']
        except Exception as e:
            logger.debug("Collection %s failing with %s" % (collectionName, str(e)))
        return sf

    def getDocumentKeyAttributeNames(self, collectionName):
        r = []
        try:
            return self.__cfgD['collection_attribute_names'][collectionName]
        except Exception as e:
            logger.debug("Collection %s failing with %s" % (collectionName, str(e)))
        return r

    def getPrivateDocumentAttributes(self, collectionName):
        r = []
        try:
            return [d for d in self.__cfgD['collection_private_keys'][collectionName] if d['PRIVATE_DOCUMENT_NAME'] and len(d['PRIVATE_DOCUMENT_NAME']) > 0]
        except Exception as e:
            logger.debug("Collection %s failing with %s" % (collectionName, str(e)))
        return r

    def getSubCategoryAggregates(self, collectionName):
        r = []
        try:
            return self.__cfgD['collection_subcategory_aggregates'][collectionName]
        except Exception as e:
            logger.debug("Collection %s failing with %s" % (collectionName, str(e)))
        return r



##
# File:    SchemaDefBuild.py
# Author:  J. Westbrook
# Date:    1-May-2018
# Version: 0.001 Initial version
#
# Updates:
#
#  9-May-2018 jdw integrate dictionary and file based (type/coverage) data.
#  7-Aug-2018 jdw add slice definitions converted to schema id references
# 13-Aug-2018 jdw Refine the role of includeContentClasses -
# 14-Aug-2018 jdw Return 'COLLECTION_DOCUMENT_ATTRIBUTE_NAMES' as a list
#  6-Sep-2018 jdw Generalize JSON schema generation method
# 14-Sep-2018 jdw Require at least one record in any array type, adjust constraints on iterables.
# 18-Sep-2018 jdw Constrain categories/class to homogeneous content
#  7-Oct-2018 jdw Add subCategory aggregation in the JSON schema generator
#  9-Oct-2018 jdw push the constructor arguments into the constructor as configuration options
# 12-Oct-2018 jdw filter empty required attributes in subcategory aggregates
# 24-Oct-2018 jdw update for new configuration organization
# 18-Nov-2018 jdw add COLLECTION_DOCUMENT_ATTRIBUTE_INFO
##
"""
Integrate dictionary metadata and file based(type/coverage) into internal and JSON/BSON schema defintions.


"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import copy
import logging

from rcsb.db.define.DataTypeApplicationInfo import DataTypeApplicationInfo
from rcsb.db.define.DataTypeInstanceInfo import DataTypeInstanceInfo
from rcsb.db.define.DictInfo import DictInfo
from rcsb.utils.config.ConfigUtil import ConfigUtil

logger = logging.getLogger(__name__)


class SchemaDefBuild(object):
    """ Integrate dictionary metadata and file based(type/coverage) into internal and JSON/BSON schema defintions.

    """

    def __init__(self, schemaName, configPath, configName='site_info', mockTopPath=None, includeContentClasses=['GENERATED_CONTENT', 'EVOLVING_CONTENT']):
        """

        """
        self.__cfgOb = ConfigUtil(configPath=configPath, defaultSectionName=configName, mockTopPath=mockTopPath)
        self.__schemaName = schemaName
        self.__includeContentClasses = includeContentClasses
        #
        pathPdbxDictionaryFile = self.__cfgOb.getPath('PDBX_DICT_LOCATOR', sectionName=configName)
        pathRcsbDictionaryFile = self.__cfgOb.getPath('RCSB_DICT_LOCATOR', sectionName=configName)
        dictLocators = [pathPdbxDictionaryFile, pathRcsbDictionaryFile]
        #
        self.__instDataTypeFilePath = self.__cfgOb.getPath('INSTANCE_DATA_TYPE_INFO_LOCATOR', sectionName=schemaName)
        self.__appDataTypeFilePath = self.__cfgOb.getPath('APP_DATA_TYPE_INFO_LOCATOR', sectionName=configName)

        dictHelper = self.__cfgOb.getHelper('DICT_HELPER_MODULE', sectionName=configName, cfgOb=self.__cfgOb)
        self.__schemaDefHelper = self.__cfgOb.getHelper('SCHEMADEF_HELPER_MODULE', sectionName=configName, cfgOb=self.__cfgOb)
        self.__documentDefHelper = self.__cfgOb.getHelper('DOCUMENT_HELPER_MODULE', sectionName=configName, cfgOb=self.__cfgOb)

        self.__dictInfo = DictInfo(dictLocators=dictLocators, dictHelper=dictHelper, dictSubset=schemaName)
        #

    def build(self, collectionName=None, applicationName='ANY', schemaType='rcsb', enforceOpts="mandatoryKeys|mandatoryAttributes|bounds|enums"):
        rD = {}
        if schemaType.lower() == 'rcsb':
            rD = self.__build(schemaName=self.__schemaName,
                              applicationName=applicationName,
                              instDataTypeFilePath=self.__instDataTypeFilePath,
                              appDataTypeFilePath=self.__appDataTypeFilePath,
                              schemaDefHelper=self.__schemaDefHelper,
                              documentDefHelper=self.__documentDefHelper,
                              includeContentClasses=self.__includeContentClasses
                              )
        elif schemaType.lower() in ['json', 'bson']:
            rD = self.__createJsonLikeSchema(schemaName=self.__schemaName,
                                             collectionName=collectionName,
                                             applicationName=applicationName.upper(),
                                             instDataTypeFilePath=self.__instDataTypeFilePath,
                                             appDataTypeFilePath=self.__appDataTypeFilePath,
                                             schemaDefHelper=self.__schemaDefHelper,
                                             documentDefHelper=self.__documentDefHelper,
                                             includeContentClasses=self.__includeContentClasses,
                                             enforceOpts=enforceOpts
                                             )

        return rD

    def __build(self, schemaName, applicationName, instDataTypeFilePath, appDataTypeFilePath,
                schemaDefHelper, documentDefHelper, includeContentClasses):
        """
        """
        databaseName = self.__schemaDefHelper.getDatabaseName(schemaName) if self.__schemaDefHelper else ''
        databaseVersion = self.__schemaDefHelper.getDatabaseVersion(schemaName) if self.__schemaDefHelper else ''
        #
        schemaDef = {'NAME': schemaName, 'APP_NAME': applicationName, 'DATABASE_NAME': databaseName, 'DATABASE_VERSION': databaseVersion}
        #
        schemaDef['SELECTION_FILTERS'] = self.__dictInfo.getSelectionFiltersForSubset()

        schemaDef['SCHEMA_DICT'] = self.__createSchemaDict(schemaName, applicationName, instDataTypeFilePath,
                                                           appDataTypeFilePath, schemaDefHelper, includeContentClasses)
        schemaDef['DOCUMENT_DICT'] = self.__createDocumentDict(schemaName, documentDefHelper)
        schemaDef['SLICE_PARENT_ITEMS'] = self.__convertSliceParentItemNames(schemaName, applicationName)
        schemaDef['SLICE_PARENT_FILTERS'] = self.__convertSliceParentFilterNames(schemaName, applicationName)
        return schemaDef

    def __createDocumentDict(self, schemaName, documentDefHelper, applicationName='ANY'):
        """Internal method to assign document-level details to the schema definition,


        Args:
            schemaName (string): A schema/content name: pdbx|chem_comp|bird|bird_family ...
            documentDefHelper (class instance):  Class instance providing additional document-level metadata
            applicationName (string, optional): Application name ANY|SQL|...

        Returns:
            dict: dictionary of document-level metadata


        """
        rD = {'CONTENT_TYPE_COLLECTION_MAP': {},
              'COLLECTION_DOCUMENT_ATTRIBUTE_NAMES': {},
              'COLLECTION_DOCUMENT_PRIVATE_KEYS': {},
              'COLLECTION_CONTENT': {}}
        #
        dH = documentDefHelper
        if dH:
            cL = dH.getCollections(schemaName)
            rD['CONTENT_TYPE_COLLECTION_MAP'][schemaName] = cL
            for c in cL:
                rD['COLLECTION_CONTENT'][c] = {'INCLUDE': dH.getIncluded(c), 'EXCLUDE': dH.getExcluded(c), 'SLICE_FILTER': dH.getSliceFilter(c)}
                rD['COLLECTION_DOCUMENT_ATTRIBUTE_NAMES'][c] = dH.getDocumentKeyAttributeNames(c)
                rD['COLLECTION_DOCUMENT_PRIVATE_KEYS'][c] = dH.getPrivateDocumentAttributes(c)
        #
        return rD

    def __testContentClasses(self, includeContentClasses, assignedContentClasses):
        """ Return True if any of the include content classes are assigned.

        """
        # logger.debug("includeContentClasses %r assignedContentClasses %r" % (includeContentClasses, assignedContentClasses))
        for cc in includeContentClasses:
            if cc in assignedContentClasses:
                return True
        return False

    def __getConvertNameMethod(self, applicationName):
        # Function to perform category and attribute name conversion.
        # convertNameF = self.__schemaDefHelper.convertNameDefault if self.__schemaDefHelper else self.__convertNameDefault
        #
        try:
            if applicationName in ['ANY', 'SQL', 'DOCUMENT', 'SOLR', 'JSON', 'BSON']:
                nameConvention = applicationName
            else:
                nameConvention = 'DEFAULT'
            return self.__schemaDefHelper.getConvertNameMethod(nameConvention) if self.__schemaDefHelper else self.__convertNameDefault
        except Exception:
            pass

        return self.__convertNameDefault

    def __createSchemaDict(self, schemaName, applicationName, instDataTypeFilePath, appDataTypeFilePath, schemaDefHelper, includeContentClasses=None):
        """Internal method to integrate dictionary and instance metadata into a common schema description data structure.

        Args:
            schemaName (string): A schema/content name: pdbx|chem_comp|bird|bird_family ...
            applicationName (string): ANY|SQL
            instDataTypeFilePath (string): Path to data instance type and coverage
            appDataTypeFilePath (string): Path to resource file mapping cif data types to application data types
            schemaDefHelper (class instance): Class instance providing additional schema details
            includeContentClasses (list, optional): list of additional content classes to be included (e.g. GENERATED_CONTENT)

        Returns:
            dict: definitions for each schema object


        """
        verbose = False
        contentClasses = includeContentClasses if includeContentClasses else []
        logger.debug("Including additional category classes %r" % contentClasses)
        #
        dtInstInfo = DataTypeInstanceInfo(instDataTypeFilePath)
        dtAppInfo = DataTypeApplicationInfo(appDataTypeFilePath, applicationName=applicationName)
        #
        # Supplied by the schemaDefHelper
        #
        includeList = self.__schemaDefHelper.getIncluded(schemaName) if self.__schemaDefHelper else []
        excludeList = self.__schemaDefHelper.getExcluded(schemaName) if self.__schemaDefHelper else []
        excludeAttributesD = self.__schemaDefHelper.getExcludedAttributes(schemaName) if self.__schemaDefHelper else {}
        #
        logger.debug("Schema include list length %d" % len(includeList))
        logger.debug("Schema exclude list length %d" % len(excludeList))
        #
        # Optional synthetic attribute added to each category with value linked to data block identifier (or other function)
        #
        blockAttributeName = self.__schemaDefHelper.getBlockAttributeName(schemaName) if self.__schemaDefHelper else None
        blockAttributeCifType = self.__schemaDefHelper.getBlockAttributeCifType(schemaName) if self.__schemaDefHelper else None
        blockAttributeAppType = dtAppInfo.getAppTypeName(blockAttributeCifType)
        blockAttributeWidth = self.__schemaDefHelper.getBlockAttributeMaxWidth(schemaName) if self.__schemaDefHelper else 0
        blockAttributeMethod = self.__schemaDefHelper.getBlockAttributeMethod(schemaName) if self.__schemaDefHelper else None
        #
        convertNameF = self.__getConvertNameMethod(applicationName)
        #
        dictSchema = self.__dictInfo.getSchemaNames()
        logger.debug("Dictionary category length %d" % len(dictSchema))
        #
        rD = {}
        for catName, fullAtNameList in dictSchema.items():
            atNameList = [at for at in fullAtNameList if (catName, at) not in excludeAttributesD]
            cfD = self.__dictInfo.getCategoryFeatures(catName)
            # logger.debug("catName %s contentClasses %r cfD %r" % (catName, contentClasses, cfD))

            if not dtInstInfo.exists(catName) and not self.__testContentClasses(contentClasses, cfD['CONTENT_CLASSES']):
                logger.debug("Schema %r Skipping category %s content classes %r" % (schemaName, catName, cfD['CONTENT_CLASSES']))
                continue
            sName = convertNameF(catName)
            sId = sName.upper()
            #
            if excludeList and sId in excludeList:
                continue
            if includeList and sId not in includeList:
                continue
            # JDW
            if not cfD:
                logger.info("%s catName %s contentClasses %r cfD %r" % (schemaName, catName, contentClasses, cfD))
            #
            aD = self.__dictInfo.getAttributeFeatures(catName)
            #
            sliceNames = self.__dictInfo.getSliceNames()
            d = {}
            d['SCHEMA_ID'] = sId
            d['SCHEMA_NAME'] = sName
            d['SCHEMA_TYPE'] = 'transactional'
            d['SCHEMA_UNIT_CARDINALITY'] = cfD['UNIT_CARDINALITY'] if 'UNIT_CARDINALITY' in cfD else False
            d['SCHEMA_CONTENT_CLASSES'] = cfD['CONTENT_CLASSES'] if 'CONTENT_CLASSES' in cfD else []
            d['SCHEMA_MANDATORY'] = cfD['IS_MANDATORY']
            #
            d['ATTRIBUTES'] = {convertNameF(blockAttributeName).upper(): convertNameF(blockAttributeName)} if blockAttributeName else {}
            d['ATTRIBUTES'].update({(convertNameF(at)).upper(): convertNameF(at) for at in atNameList})
            #
            #
            d['ATTRIBUTE_MAP'] = {(convertNameF(blockAttributeName)).upper(): {'CATEGORY': None, 'ATTRIBUTE': None,
                                                                               'METHOD_NAME': blockAttributeMethod, 'ARGUMENTS': None}} if blockAttributeName else {}

            d['ATTRIBUTE_INFO'] = {}
            atIdIndexList = []
            atNameIndexList = []
            iOrder = 1
            if blockAttributeName:
                td = {
                    'ORDER': iOrder,
                    'NULLABLE': False,
                    'PRECISION': 0,
                    'PRIMARY_KEY': True,
                    'APP_TYPE': blockAttributeAppType,
                    'WIDTH': blockAttributeWidth,
                    'ITERABLE_DELIMITER': None,
                    'FILTER_TYPES': [],
                    'ENUMERATION': {},
                    'IS_CHAR_TYPE': True,
                    'CONTENT_CLASSES': ['BLOCK_ATTRIBUTE']}
                iOrder += 1
                atId = (convertNameF(blockAttributeName)).upper()
                atIdIndexList.append(atId)
                atNameIndexList.append(blockAttributeName)
                d['ATTRIBUTE_INFO'][atId] = td
            #
            for atName in sorted(atNameList):
                fD = aD[atName]
                if not dtInstInfo.exists(catName, atName) and not self.__testContentClasses(contentClasses, fD['CONTENT_CLASSES']):
                    continue
                if fD['IS_KEY']:
                    appType = dtAppInfo.getAppTypeName(fD['TYPE_CODE'])
                    appWidth = dtAppInfo.getAppTypeDefaultWidth(fD['TYPE_CODE'])
                    instWidth = dtInstInfo.getMaxWidth(catName, atName)
                    #
                    revAppType, revAppWidth = dtAppInfo.updateCharType(fD['IS_KEY'], appType, instWidth, appWidth, bufferPercent=20.0)
                    if verbose and applicationName in ['SQL', 'ANY']:
                        logger.debug("catName %s atName %s cifType %s appType %s appWidth %r instWidth %r --> revAppType %r revAppWidth %r " %
                                     (catName, atName, fD['TYPE_CODE'], appType, appWidth, instWidth, revAppType, revAppWidth))
                    #
                    appPrecision = dtAppInfo.getAppTypeDefaultPrecision(fD['TYPE_CODE'])
                    td = {'ORDER': iOrder,
                          'NULLABLE': not fD['IS_MANDATORY'],
                          'PRECISION': appPrecision,
                          'PRIMARY_KEY': fD['IS_KEY'],
                          'APP_TYPE': revAppType,
                          'WIDTH': revAppWidth,
                          'ITERABLE_DELIMITER': None,
                          'FILTER_TYPES': fD['FILTER_TYPES'],
                          'IS_CHAR_TYPE': fD['IS_CHAR_TYPE'],
                          'ENUMERATION': {str(ky).lower(): ky for ky in fD['ENUMS']},
                          'CONTENT_CLASSES': fD['CONTENT_CLASSES']}
                    atId = (convertNameF(atName)).upper()
                    d['ATTRIBUTE_INFO'][atId] = td
                    atIdIndexList.append(atId)
                    atNameIndexList.append(atName)
                    #
                    mI = self.__dictInfo.getMethodImplementation(catName, atName, methodCodes=["calculate_on_load"])
                    if mI:
                        d['ATTRIBUTE_MAP'].update({(convertNameF(atName)).upper(): {'CATEGORY': None, 'ATTRIBUTE': None, 'METHOD_NAME': mI, 'ARGUMENTS': None}})
                    else:
                        d['ATTRIBUTE_MAP'].update({(convertNameF(atName)).upper(): {'CATEGORY': catName, 'ATTRIBUTE': atName, 'METHOD_NAME': None, 'ARGUMENTS': None}})
                    iOrder += 1
            for atName in sorted(atNameList):
                fD = aD[atName]
                if not dtInstInfo.exists(catName, atName) and not self.__testContentClasses(contentClasses, fD['CONTENT_CLASSES']):
                    continue

                if not fD['IS_KEY']:
                    appType = dtAppInfo.getAppTypeName(fD['TYPE_CODE'])
                    appWidth = dtAppInfo.getAppTypeDefaultWidth(fD['TYPE_CODE'])
                    instWidth = dtInstInfo.getMaxWidth(catName, atName)
                    revAppType, revAppWidth = dtAppInfo.updateCharType(fD['IS_KEY'], appType, instWidth, appWidth, bufferPercent=20.0)
                    if verbose and applicationName in ['SQL', 'ANY']:
                        logger.debug("catName %s atName %s cifType %s appType %s appWidth %r instWidth %r --> revAppType %r revAppWidth %r " %
                                     (catName, atName, fD['TYPE_CODE'], appType, appWidth, instWidth, revAppType, revAppWidth))

                    #
                    appPrecision = dtAppInfo.getAppTypeDefaultPrecision(fD['TYPE_CODE'])
                    td = {'ORDER': iOrder,
                          'NULLABLE': not fD['IS_MANDATORY'],
                          'PRECISION': appPrecision,
                          'PRIMARY_KEY': fD['IS_KEY'],
                          'APP_TYPE': revAppType,
                          'WIDTH': revAppWidth,
                          'ITERABLE_DELIMITER': fD['ITERABLE_DELIMITER'],
                          'FILTER_TYPES': fD['FILTER_TYPES'],
                          'IS_CHAR_TYPE': fD['IS_CHAR_TYPE'],
                          'ENUMERATION': {str(ky).lower(): ky for ky in fD['ENUMS']},
                          'CONTENT_CLASSES': fD['CONTENT_CLASSES']}
                    atId = (convertNameF(atName)).upper()
                    d['ATTRIBUTE_INFO'][atId] = td
                    mI = self.__dictInfo.getMethodImplementation(catName, atName, methodCodes=["calculate_on_load"])
                    if mI:
                        d['ATTRIBUTE_MAP'].update({(convertNameF(atName)).upper(): {'CATEGORY': None, 'ATTRIBUTE': None, 'METHOD_NAME': mI, 'ARGUMENTS': None}})
                    else:
                        d['ATTRIBUTE_MAP'].update({(convertNameF(atName)).upper(): {'CATEGORY': catName, 'ATTRIBUTE': atName, 'METHOD_NAME': None, 'ARGUMENTS': None}})
                    iOrder += 1
            #
            atIdDelete = convertNameF(blockAttributeName).upper() if blockAttributeName else None
            d['SCHEMA_DELETE_ATTRIBUTE'] = atIdDelete

            d['INDICES'] = {"p1": {"TYPE": "UNIQUE", "ATTRIBUTES": tuple(atIdIndexList)}}
            if len(atIdIndexList) > 1:
                d['INDICES']["s1"] = {"TYPE": "SEARCH", "ATTRIBUTES": tuple([atIdDelete])}
            #
            # JDW -  Need to review attribute names here -
            d['MAP_MERGE_INDICES'] = {catName: {'ATTRIBUTES': tuple(atNameIndexList), 'TYPE': 'EQUI-JOIN'}}
            # ----
            tD = {}
            logger.debug("Slice names %r" % sliceNames)
            for sliceName in sliceNames:
                sL = self.__dictInfo.getSliceAttributes(sliceName, catName)
                logger.debug("Slice attributes %r" % sL)
                if sL:
                    # Convert names to IDs --
                    tL = []
                    for s in sL:
                        pD = {'PARENT_CATEGORY': convertNameF(s['PARENT_CATEGORY_NAME']).upper(),
                              'PARENT_ATTRIBUTE': convertNameF(s['PARENT_ATTRIBUTE_NAME']).upper(),
                              'CHILD_ATTRIBUTE': convertNameF(s['CHILD_ATTRIBUTE_NAME']).upper()}
                        tL.append(pD)
                    tD[sliceName] = tL
            d['SLICE_ATTRIBUTES'] = tD
            #
            # ---- slice cardinality
            #
            d['SLICE_UNIT_CARDINALITY'] = {}
            sliceCardD = self.__dictInfo.getSliceUnitCardinalityForSubset()
            logger.debug("Slice card dict %r" % sliceCardD.items())
            for sliceName, catL in sliceCardD.items():
                if catName in catL:
                    d['SLICE_UNIT_CARDINALITY'][sliceName] = True
                else:
                    d['SLICE_UNIT_CARDINALITY'][sliceName] = False
            #
            d['SLICE_CATEGORY_EXTRAS'] = {}
            sliceCatD = self.__dictInfo.getSliceCategoryExtrasForSubset()
            logger.debug("Slice category extra dict %r" % sliceCatD.items())
            for sliceName, catL in sliceCatD.items():
                if catName in catL:
                    d['SLICE_CATEGORY_EXTRAS'][sliceName] = True
                else:
                    d['SLICE_CATEGORY_EXTRAS'][sliceName] = False
            #
            rD[sId] = d
        #
        return rD

    def __convertSliceParentItemNames(self, schemaName, applicationName):
        sliceD = {}
        try:
            convertNameF = self.__getConvertNameMethod(applicationName)
            # [{'CATEGORY_NAME': 'entity', 'ATTRIBUTE_NAME': 'id'}
            spD = self.__dictInfo.getSliceParentItemsForSubset()
            for ky in spD:
                rL = []
                for aL in spD[ky]:
                    d = {'CATEGORY': convertNameF(aL['CATEGORY_NAME']).upper(), 'ATTRIBUTE': convertNameF(aL['ATTRIBUTE_NAME']).upper()}
                    rL.append(d)
                sliceD[ky] = rL
            #
            return sliceD
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

        return sliceD

    def __convertSliceParentFilterNames(self, schemaName, applicationName):
        sliceD = {}
        try:
            convertNameF = self.__getConvertNameMethod(applicationName)
            # [{'CATEGORY_NAME': 'entity', 'ATTRIBUTE_NAME': 'id'}
            spD = self.__dictInfo.getSliceParentFiltersForSubset()
            for ky in spD:
                rL = []
                for aL in spD[ky]:
                    d = {'CATEGORY': convertNameF(aL['CATEGORY_NAME']).upper(), 'ATTRIBUTE': convertNameF(aL['ATTRIBUTE_NAME']).upper(), 'VALUES': aL['VALUES']}
                    rL.append(d)
                sliceD[ky] = rL
            #
            return sliceD
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

        return sliceD

    def __convertNameDefault(self, name):
        """ Default schema name converter -
        """
        return name

    # -------------------------- ------------- ------------- ------------- ------------- ------------- -------------

    def __createJsonLikeSchema(self, schemaName, collectionName, applicationName, instDataTypeFilePath, appDataTypeFilePath,
                               schemaDefHelper, documentDefHelper, includeContentClasses=None, jsonSpecDraft='4',
                               enforceOpts="mandatoryKeys|mandatoryAttributes|bounds|enums",
                               suppressSingleton=True):
        """Internal method to integrate dictionary and instance metadata into a common json/bson schema description data structure.

           Working only for practical schema style: rowwise_by_name_with_cardinality

        Args:
            schemaName (str): A schema/content name: pdbx|chem_comp|bird|bird_family ...
            collectionName (str): Collection defined within a schema/content type
            applicationName (str): Target application for the schema (e.g. JSON, BSON, or a variant of these...)
            instDataTypeFilePath (str): Path to data instance type and coverage
            appDataTypeFilePath (str): Path to resource file mapping cif data types to application data types
            schemaDefHelper (class instance): Class instance providing additional schema details
            documentDefHelper (class instance): Class instance providing additional document schema details
            includeContentClasses (list, optional): list of additional content classes to be included (e.g. GENERATED_CONTENT)
            jsonSpecDraft (str, optional): The target draft schema specification '4|6'
            enforceOpts (str, optional): options for semantics are included in the schema (e.g. "mandatoryKeys|mandatoryAttributes|bounds|enums")

        Returns:
            dict: representation of JSON/BSON schema -


        """
        subCategoryAggregates = documentDefHelper.getSubCategoryAggregates(collectionName)
        privDocKeyL = documentDefHelper.getPrivateDocumentAttributes(collectionName)
        # enforceOpts = "mandatoryKeys|mandatoryAttributes|bounds|enums"
        #
        # applicationName = 'JSON'
        appNameU = applicationName.upper()
        typeKey = 'bsonType' if appNameU == 'BSON' else 'type'
        convertNameF = self.__getConvertNameMethod(appNameU)
        #
        addBlockAttribute = False
        contentClasses = includeContentClasses if includeContentClasses else []
        logger.debug("Including additional category classes %r" % contentClasses)
        #
        dtInstInfo = DataTypeInstanceInfo(instDataTypeFilePath)
        dtAppInfo = DataTypeApplicationInfo(appDataTypeFilePath, applicationName=appNameU)
        #
        #      Supplied by the schemaDefHelper for the content type (SchemaIds)
        #
        includeList = schemaDefHelper.getIncluded(schemaName) if self.__schemaDefHelper else []
        excludeList = schemaDefHelper.getExcluded(schemaName) if self.__schemaDefHelper else []
        excludeAttributesD = self.__schemaDefHelper.getExcludedAttributes(schemaName) if self.__schemaDefHelper else {}
        #
        #      Supplied by the documentDefHelp for the collection (SchemaIds)
        #
        docIncludeList = documentDefHelper.getIncluded(collectionName)
        docExcludeList = documentDefHelper.getExcluded(collectionName)

        sliceFilter = documentDefHelper.getSliceFilter(collectionName)
        sliceCategories = self.__dictInfo.getSliceCategories(sliceFilter) if sliceFilter else []
        sliceCategoryExtrasD = self.__dictInfo.getSliceCategoryExtrasForSubset() if sliceFilter else {}
        if sliceFilter in sliceCategoryExtrasD:
            sliceCategories.extend(sliceCategoryExtrasD[sliceFilter])
        sliceCardD = self.__dictInfo.getSliceUnitCardinalityForSubset() if sliceFilter else {}
        #
        if addBlockAttribute:
            # Optional synthetic attribute added to each category with value linked to data block identifier (or other function)
            blockAttributeName = schemaDefHelper.getBlockAttributeName(schemaName) if self.__schemaDefHelper else None
            blockAttributeCifType = schemaDefHelper.getBlockAttributeCifType(schemaName) if self.__schemaDefHelper else None
            blockAttributeAppType = dtAppInfo.getAppTypeName(blockAttributeCifType)
            blockAttributeWidth = schemaDefHelper.getBlockAttributeMaxWidth(schemaName) if schemaDefHelper else 0
            # blockAttributeMethod = schemaDefHelper.getBlockAttributeMethod(schemaName) if schemaDefHelper else None
        #
        dictSchema = self.__dictInfo.getSchemaNames()
        #
        schemaPropD = {}
        mandatoryCategoryL = []
        for catName, fullAtNameList in dictSchema.items():
            atNameList = [at for at in fullAtNameList if (catName, at) not in excludeAttributesD]
            cfD = self.__dictInfo.getCategoryFeatures(catName)
            # logger.debug("catName %s contentClasses %r cfD %r" % (catName, contentClasses, cfD))

            #
            #  Skip categories that are uniformly unpopulated --
            #
            if not dtInstInfo.exists(catName) and not self.__testContentClasses(contentClasses, cfD['CONTENT_CLASSES']):
                logger.debug("Schema %r Skipping category %s content classes %r" % (schemaName, catName, cfD['CONTENT_CLASSES']))
                continue
            #
            # -> Create a schema id  for catName <-
            sName = convertNameF(catName)
            schemaId = sName.upper()
            #
            #  These are the content type schema level filters -
            if excludeList and schemaId in excludeList:
                continue
            if includeList and schemaId not in includeList:
                continue
            #
            # These are collection level filters
            #
            if docExcludeList and schemaId in docExcludeList:
                continue
            if docIncludeList and schemaId not in docIncludeList:
                continue
            #
            #  If there is a slice filter on this collection, the skip categories not connected to the slice
            if sliceFilter and catName not in sliceCategories:
                continue
            #
            #        Done with category filtering/selections
            # -------- ---------- ------------ -------- ---------- ------------ -------- ---------- ------------
            #
            aD = self.__dictInfo.getAttributeFeatures(catName)
            #
            if cfD['IS_MANDATORY']:
                mandatoryCategoryL.append(catName)
            #
            isUnitCard = True if ('UNIT_CARDINALITY' in cfD and cfD['UNIT_CARDINALITY']) else False
            if sliceFilter and sliceFilter in sliceCardD:
                isUnitCard = catName in sliceCardD[sliceFilter]
            #
            pD = {typeKey: "object", 'properties': {}, 'required': []}
            #
            if isUnitCard:
                catPropD = pD
            else:
                if cfD['IS_MANDATORY']:
                    # catPropD = {typeKey: "array", 'items': [pD], 'minItems': 1, 'uniqueItems': True}
                    catPropD = {typeKey: "array", 'items': pD, 'minItems': 1, 'uniqueItems': True}
                else:
                    # JDW Adjusted minItems=1
                    catPropD = {typeKey: "array", 'items': pD, 'minItems': 1, 'uniqueItems': True}
            #
            if addBlockAttribute:
                schemaAttributeName = convertNameF(blockAttributeName)
                atPropD = {typeKey: blockAttributeAppType, 'maxWidth': blockAttributeWidth}
                pD['required'].append(schemaAttributeName)
                pD['properties'][schemaAttributeName] = atPropD

            #  First, filter any subcategory aggregates from the available list of a category attributes
            #
            subCatPropD = {}
            if subCategoryAggregates:
                for subCategory in subCategoryAggregates:
                    reqL = []
                    scD = {typeKey: "object", 'properties': {}, }
                    for atName in sorted(atNameList):
                        fD = aD[atName]
                        # Exclude primary data attributes with no instance coverage except if in a protected content class
                        if not dtInstInfo.exists(catName, atName) and not self.__testContentClasses(contentClasses, fD['CONTENT_CLASSES']):
                            continue
                        if subCategory not in fD['SUB_CATEGORIES']:
                            continue
                        #
                        schemaAttributeName = convertNameF(atName)
                        isRequired = ('mandatoryAttributes' in enforceOpts and fD['IS_MANDATORY'])
                        if isRequired:
                            reqL.append(schemaAttributeName)
                        #
                        atPropD = self.__getJsonAttributeProperties(fD, appNameU, dtAppInfo, jsonSpecDraft, enforceOpts)
                        scD['properties'][schemaAttributeName] = atPropD
                    if reqL:
                        scD['required'] = reqL
                subCatPropD[subCategory] = {typeKey: 'array', 'items': scD, 'uniqueItems': True}
            #
            if subCatPropD:
                logger.debug("subCatPropD %r" % subCatPropD.items())
            #
            for atName in sorted(atNameList):
                fD = aD[atName]
                # Exclude primary data attributes with no instance coverage except if in a protected content class
                if not dtInstInfo.exists(catName, atName) and not self.__testContentClasses(contentClasses, fD['CONTENT_CLASSES']):
                    continue
                if subCategoryAggregates and self.__subCategoryTest(subCategoryAggregates, fD['SUB_CATEGORIES']):
                    continue
                #
                schemaAttributeName = convertNameF(atName)
                isRequired = (('mandatoryKeys' in enforceOpts and fD['IS_KEY']) or ('mandatoryAttributes' in enforceOpts and fD['IS_MANDATORY']))
                if isRequired:
                    pD['required'].append(schemaAttributeName)
                #
                atPropD = self.__getJsonAttributeProperties(fD, appNameU, dtAppInfo, jsonSpecDraft, enforceOpts)

                delimiter = fD['ITERABLE_DELIMITER']
                if delimiter:
                    pD['properties'][schemaAttributeName] = {typeKey: 'array', 'items': atPropD, 'uniqueItems': False}
                else:
                    pD['properties'][schemaAttributeName] = atPropD

            pD['properties'].update(subCatPropD)
            pD['required'].extend(list(subCatPropD.keys()))
            #
            if 'required' in catPropD and len(catPropD['required']) < 1:
                logger.info("Category %s cfD %r" % (catName, cfD.items()))
            #
            schemaPropD[sName] = copy.deepcopy(catPropD)
        #
        # Add any private keys to the object schema - Fetch the metadata for the private keys
        #
        privKeyD = {}
        if privDocKeyL:
            for pdk in privDocKeyL:
                aD = self.__dictInfo.getAttributeFeatures(convertNameF(pdk['CATEGORY_NAME']))
                fD = aD[convertNameF(pdk['ATTRIBUTE_NAME'])]
                atPropD = self.__getJsonAttributeProperties(fD, appNameU, dtAppInfo, jsonSpecDraft, enforceOpts)
                privKeyD[pdk['PRIVATE_DOCUMENT_NAME']] = atPropD

        #
        # Suppress the category name for schemas with a single category -
        #
        if suppressSingleton and len(schemaPropD) == 1:
            logger.debug("%s %s suppressing category in singleton schema" % (schemaName, collectionName))
            # rD = copy.deepcopy(catPropD)
            for k, v in privKeyD.items():
                pD['properties'][k] = v
                pD['required'] = k
            rD = copy.deepcopy(pD)

        else:
            for k, v in privKeyD.items():
                schemaPropD[k] = v
                mandatoryCategoryL.append(k)
            #
            rD = {typeKey: 'object', 'properties': schemaPropD}
            if len(mandatoryCategoryL):
                rD['required'] = mandatoryCategoryL

        if appNameU == 'JSON':
            jsonSchemaUrl = "http://json-schema.org/draft-0%s/schema#" % jsonSpecDraft if jsonSpecDraft in ['3', '4', '6', '7'] else "http://json-schema.org/schema#"
            rD.update({"$id": "https://github.com/rcsb/py-rcsb.db/tree/master/rcsb.db/data/json-schema/",
                       "$schema": jsonSchemaUrl,
                       'title': 'Schema for content type %s collection %s' % (schemaName, collectionName)})

        return rD

    def __getJsonAttributeProperties(self, fD, appNameU, dtAppInfo, jsonSpecDraft, enforceOpts):
        #
        atPropD = {}
        try:
            # - assign data type attributes
            typeKey = 'bsonType' if appNameU == 'BSON' else 'type'
            appType = dtAppInfo.getAppTypeName(fD['TYPE_CODE'])
            #
            #
            if appType in ['string']:
                # atPropD = {typeKey: appType, 'maxWidth': instWidth}
                atPropD = {typeKey: appType}
            elif appType in ['date', 'datetime'] and appNameU == 'JSON':
                fmt = 'date' if appType == 'date' else 'date-time'
                atPropD = {typeKey: 'string', 'format': fmt}
            elif appType in ['date', 'datetime'] and appNameU == 'BSON':
                atPropD = {typeKey: 'date'}
            elif appType in ['number', 'integer', 'int', 'double']:
                atPropD = {typeKey: appType}
                #
                if 'bounds' in enforceOpts:
                    if jsonSpecDraft in ['3', '4']:
                        if 'MIN_VALUE' in fD:
                            atPropD['minimum'] = fD['MIN_VALUE']
                        elif 'MIN_VALUE_EXCLUSIVE' in fD:
                            atPropD['minimum'] = fD['MIN_VALUE_EXCLUSIVE']
                            atPropD['exclusiveMinimum'] = True
                        if 'MAX_VALUE' in fD:
                            atPropD['maximum'] = fD['MAX_VALUE']
                        elif 'MAX_VALUE_EXCLUSIVE' in fD:
                            atPropD['maximum'] = fD['MAX_VALUE_EXCLUSIVE']
                            atPropD['exclusiveMaximum'] = True
                    elif jsonSpecDraft in ['6', '7']:
                        if 'MIN_VALUE' in fD:
                            atPropD['minimum'] = fD['MIN_VALUE']
                        elif 'MIN_VALUE_EXCLUSIVE' in fD:
                            atPropD['exclusiveMinimum'] = fD['MIN_VALUE_EXCLUSIVE']
                        if 'MAX_VALUE' in fD:
                            atPropD['maximum'] = fD['MAX_VALUE']
                        elif 'MAX_VALUE_EXCLUSIVE' in fD:
                            atPropD['exclusiveMaximum'] = fD['MAX_VALUE_EXCLUSIVE']
            else:
                atPropD = {typeKey: appType}
            #
            if 'enums' in enforceOpts and fD['ENUMS']:
                atPropD['enum'] = fD['ENUMS']
            if appNameU not in ['BSON']:
                try:
                    if fD['EXAMPLES']:
                        atPropD['examples'] = [str(t1).strip() for t1, t2 in fD['EXAMPLES']]
                except Exception as e:
                    logger.exception("Failing for %r with %s" % (fD['EXAMPLES'], str(e)))
                if fD['DESCRIPTION']:
                    atPropD['description'] = fD['DESCRIPTION']
                #
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        #
        return atPropD

    def __subCategoryTest(self, filterList, atSubCategoryList):
        """ Return true if any element of filter list in atSubCategoryList
        """
        if not filterList or not atSubCategoryList:
            return False
        for subCat in filterList:
            if subCat in atSubCategoryList:
                return True
        return False



##
# File: SchemaUpdateExec.py
# Date: 11-Nov-2018  jdw
#
#  Execution wrapper  --  for schema production utilities --
#
#  Updates:
#
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import argparse
import logging
import os

from rcsb.db.define.SchemaDefBuild import SchemaDefBuild
from rcsb.utils.config.ConfigUtil import ConfigUtil
from rcsb.utils.io.IoUtil import IoUtil

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format='%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s')
logger = logging.getLogger()


class SchemaUpdateWorker(object):

    def __init__(self, configPath, schemaLevel='full', schemaDirPath=None, mockTopPath=None):
        self.__configPath = configPath
        self.__schemaDirPath = schemaDirPath
        self.__mockTopPath = mockTopPath
        self.__schemaLevel = schemaLevel
        #
        if schemaLevel == 'full':
            self.__enforceOpts = "mandatoryKeys|mandatoryAttributes|bounds|enums"
        else:
            self.__enforceOpts = "mandatoryKeys|enums"

    def buildRcsbSchema(self, schemaName):
        try:
            #
            smb = SchemaDefBuild(schemaName, self.__configPath, mockTopPath=self.__mockTopPath)
            for applicationName in ['ANY', 'SQL']:
                if self.__schemaDirPath:
                    pathSchemaDefJson = os.path.join(self.__schemaDirPath, 'schema_def-%s-%s.json' % (schemaName, applicationName))
                elif self.__mockTopPath:
                    pathSchemaDefJson = os.path.join(self.__mockTopPath, 'schema', 'schema_def-%s-%s.json' % (schemaName, applicationName))
                else:
                    logger.error("No RCSB schema path can be defined for %s" % schemaName)
                    return False
                cD = smb.build(applicationName=applicationName, schemaType='rcsb')
                #
                ioU = IoUtil()
                ioU.serialize(pathSchemaDefJson, cD, format='json', indent=3)
            return True
        except Exception as e:
            logger.exception("rcsb schema generation for %r failing with %s" % (schemaName, str(e)))
        return False

    def buildJsonSchema(self, schemaName, collectionName):
        try:
            #
            if self.__schemaDirPath:
                pathSchemaDefJson = os.path.join(self.__schemaDirPath, 'json-schema-%s-%s.json' % (self.__schemaLevel, collectionName))
            elif self.__mockTopPath:
                pathSchemaDefJson = os.path.join(self.__mockTopPath, 'json', 'json-schema-%s-%s.json' % (self.__schemaLevel, collectionName))
            else:
                logger.error("No JSON schema path defined for %s" % schemaName)
                return False
            #
            logger.info("Building schema %s %s in %s" % (schemaName, collectionName, pathSchemaDefJson))
            #
            smb = SchemaDefBuild(schemaName, self.__configPath, mockTopPath=self.__mockTopPath)
            cD = smb.build(collectionName, applicationName='JSON', schemaType='JSON', enforceOpts=self.__enforceOpts)
            #
            ioU = IoUtil()
            ioU.serialize(pathSchemaDefJson, cD, format='json', indent=3)
            return True
        except Exception as e:
            logger.exception("%r %r failing with %s" % (schemaName, collectionName, str(e)))
        return False

    def buildBsonSchema(self, schemaName, collectionName):
        try:

            if self.__schemaDirPath:
                pathSchemaDefBson = os.path.join(self.__schemaDirPath, 'bson-schema-%s-%s.json' % (self.__schemaLevel, collectionName))
            elif self.__mockTopPath:
                pathSchemaDefBson = os.path.join(self.__mockTopPath, 'json', 'bson-schema-%s-%s.json' % (self.__schemaLevel, collectionName))
            else:
                logger.error("No JSON schema path defined for %s" % schemaName)
                return False
            #
            smb = SchemaDefBuild(schemaName, self.__configPath, mockTopPath=self.__mockTopPath)
            cD = smb.build(collectionName, applicationName='BSON', schemaType='BSON', enforceOpts=self.__enforceOpts)
            #
            ioU = IoUtil()
            ioU.serialize(pathSchemaDefBson, cD, format='json', indent=3)

        except Exception as e:
            logger.exception("%r %r failing with %s" % (schemaName, collectionName, str(e)))
        return False


def main():
    parser = argparse.ArgumentParser()
    #
    #
    parser.add_argument("--update_chem_comp_ref", default=False, action='store_true', help="Update schema for Chemical Component reference definitions")
    parser.add_argument("--update_chem_comp_core_ref", default=False, action='store_true', help="Update core schema for Chemical Component reference definitions")
    parser.add_argument("--update_bird_chem_comp_ref", default=False, action='store_true', help="Update schema for Bird Chemical Component reference definitions")
    parser.add_argument("--update_bird_chem_comp_core_ref", default=False, action='store_true', help="Update core schema for Bird Chemical Component reference definitions")

    parser.add_argument("--update_bird_ref", default=False, action='store_true', help="Update schema for Bird reference definitions")
    parser.add_argument("--update_bird_family_ref", default=False, action='store_true', help="Update schema for Bird Family reference definitions")

    parser.add_argument("--update_pdbx", default=False, action='store_true', help="Update schema for PDBx entry data")
    parser.add_argument("--update_pdbx_core", default=False, action='store_true', help="Update schema for PDBx core entry/entity data")
    #
    parser.add_argument("--update_repository_holdings", default=False, action='store_true', help="Update schema for repository holdings")
    parser.add_argument("--update_entity_sequence_clusters", default=False, action='store_true', help="Update schema for entity sequence clusters")
    parser.add_argument("--update_data_exchange", default=False, action='store_true', help="Update schema for data exchange status")
    #
    parser.add_argument("--config_path", default=None, help="Path to configuration options file")
    parser.add_argument("--config_name", default="site_info", help="Configuration section name")
    #
    parser.add_argument("--schema_dirpath", default=None, help="Output schema directory path")
    parser.add_argument("--schema_format", default=None, help="Schema encoding (rcsb|json|bson)")
    parser.add_argument("--schema_level", default=None, help="Schema validation level (full|min default=None)")
    #
    parser.add_argument("--debug", default=False, action='store_true', help="Turn on verbose logging")
    parser.add_argument("--mock", default=False, action='store_true', help="Use MOCK repository configuration for dependencies and testing")
    parser.add_argument("--working_path", default=None, help="Working path for temporary files")
    args = parser.parse_args()
    #
    debugFlag = args.debug
    if debugFlag:
        logger.setLevel(logging.DEBUG)
    # ----------------------- - ----------------------- - ----------------------- - ----------------------- - ----------------------- -
    #                                       Configuration Details
    configPath = args.config_path
    configName = args.config_name
    schemaDirPath = args.schema_dirpath
    #
    schemaFormat = args.schema_format
    schemaLevel = args.schema_level if args.schema_level in ['min', 'full'] else None

    if not configPath:
        configPath = os.getenv('DBLOAD_CONFIG_PATH', None)
    try:
        if os.access(configPath, os.R_OK):
            os.environ['DBLOAD_CONFIG_PATH'] = configPath
            logger.info("Using configuation path %s (%s)" % (configPath, configName))
        else:
            logger.error("Missing or access issue with config file %r" % configPath)
            exit(1)
        mockTopPath = os.path.join(TOPDIR, 'rcsb', 'mock-data') if args.mock else None
        cfgOb = ConfigUtil(configPath=configPath, defaultSectionName=configName, mockTopPath=mockTopPath)
    except Exception as e:
        logger.error("Missing or access issue with config file %r with %s" % (configPath, str(e)))
        exit(1)
    #
    suw = SchemaUpdateWorker(configPath, schemaLevel, schemaDirPath=schemaDirPath, mockTopPath=mockTopPath)
    #
    schemaNameList = []
    if args.update_chem_comp_ref:
        schemaNameList.append('chem_comp')

    if args.update_bird_chem_comp_ref:
        schemaNameList.append('bird_chem_comp')

    if args.update_chem_comp_core_ref:
        schemaNameList.append('chem_comp_core')

    if args.update_bird_chem_comp_core_ref:
        schemaNameList.append('bird_chem_comp_core')

    if args.update_bird_ref:
        schemaNameList.append('bird')

    if args.update_bird_family_ref:
        schemaNameList.append('bird_family')

    if args.update_pdbx:
        schemaNameList.append('pdbx')

    if args.update_pdbx_core:
        schemaNameList.append('pdbx_core')

    if args.update_repository_holdings:
        schemaNameList.append('repository_holdings')

    if args.update_entity_sequence_clusters:
        schemaNameList.append('entity_sequence_clusters')

    if args.update_data_exchange:
        schemaNameList.append('data_exchange')

    scnD = cfgOb.get('schema_collection_names', sectionName='document_helper')
    #
    logger.debug("Collections %s" % (list(scnD.items())))
    logger.debug("schemaNameList %s" % schemaNameList)

    for schemaName in schemaNameList:

        if schemaFormat == 'rcsb':
            logger.info("Creating RCSB schema for content type %s" % schemaName)
            suw.buildRcsbSchema(schemaName)
        if schemaName in scnD:
            for collectionName in scnD[schemaName]:
                logger.info("Creating Exchange JSON schema for content type %s %s" % (schemaName, collectionName))
                if schemaFormat == 'json':
                    logger.info("Creating Exchange JSON schema for content type %s %s" % (schemaName, collectionName))
                    suw.buildJsonSchema(schemaName, collectionName)
                elif schemaFormat == 'bson':
                    logger.info("Creating Exchange BSON schema for content type %s %s" % (schemaName, collectionName))
                    suw.buildBsonSchema(schemaName, collectionName)


if __name__ == '__main__':
    main()



##
# File:    DictMethodRunnerHelperBase.py
# Author:  J. Westbrook
# Date:    18-Aug-2018
# Version: 0.001 Initial version
#
# Updates:
##
"""
Foundation class to invoke dictioary methods implemented as helper methods

Placeholder for future developer -

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

logger = logging.getLogger(__name__)


class DictMethodRunnerHelperBase(object):
    """ Foundation class to invoke dictioary methods implemented as helper methods.

    """

    def __init__(self, **kwargs):
        self._raiseExceptions = kwargs.get('raiseExceptions', False)

    def doSomething(self, kind):
        """ Placeholder for future development
        """
        if self._raiseExceptions:
            raise NotImplementedError()
        return []



##
# File: RepoScanExec.py
#
#  Execution wrapper  --  repository scanning utilities --
#
#  Updates:
#
# 28-Jun-2018 jdw update ScanRepoUtil prototype with workPath
#  3-Jul-2018 jdw update to latest ScanRepoUtil() prototype
# 20-Aug-2018 jdw engage incremental repository scan mode.
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import argparse
import logging
import os
import sys

from rcsb.db.define.DictInfo import DictInfo
from rcsb.db.utils.ScanRepoUtil import ScanRepoUtil
from rcsb.utils.config.ConfigUtil import ConfigUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format='%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s')
logger = logging.getLogger()


def scanRepo(cfgOb, contentType, scanDataFilePath, dictFilePath, numProc, chunkSize, fileLimit, scanType='full',
             inputPathList=None, pathListFilePath=None, dataCoverageFilePath=None, dataTypeFilePath=None,
             failedFilePath=None, workPath=None):
    """ Utility method to scan the data repository of the input content type and store type and coverage details.
    """
    try:
        #
        #
        dI = DictInfo(dictLocators=[dictFilePath])
        attributeDataTypeD = dI.getAttributeDataTypeD()
        #
        sr = ScanRepoUtil(cfgOb, attributeDataTypeD=attributeDataTypeD, numProc=numProc, chunkSize=chunkSize, fileLimit=fileLimit, workPath=workPath)
        ok = sr.scanContentType(contentType, scanType=scanType, inputPathList=inputPathList, scanDataFilePath=scanDataFilePath,
                                failedFilePath=failedFilePath, saveInputFileListPath=pathListFilePath)
        if dataTypeFilePath:
            ok = sr.evalScan(scanDataFilePath, dataTypeFilePath, evalType='data_type')
        if dataCoverageFilePath:
            ok = sr.evalScan(scanDataFilePath, dataCoverageFilePath, evalType='data_coverage')

        return ok
    except Exception as e:
        logger.exception("Failing with %s" % str(e))


def main():
    parser = argparse.ArgumentParser()
    #
    parser.add_argument("--dict_file_path", default=None, help="PDBx/mmCIF dictionary file path")
    #
    parser.add_argument("--scanType", default='full', help="Repository scan type (full|incr)")
    #
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--scan_chem_comp_ref", default=False, action='store_true', help="Scan Chemical Component reference definitions (public subset)")
    group.add_argument("--scan_chem_comp_core_ref", default=False, action='store_true', help="Scan Chemical Component Core reference definitions (public subset)")
    group.add_argument("--scan_bird_chem_comp_ref", default=False, action='store_true', help="Scan Bird Chemical Component reference definitions (public subset)")
    group.add_argument("--scan_bird_chem_comp_core_ref", default=False, action='store_true', help="Scan Bird Chemical Component Core reference definitions (public subset)")
    group.add_argument("--scan_bird_ref", default=False, action='store_true', help="Scan Bird reference definitions (public subset)")
    group.add_argument("--scan_bird_family_ref", default=False, action='store_true', help="Scan Bird Family reference definitions (public subset)")
    group.add_argument("--scan_entry_data", default=False, action='store_true', help="Scan PDB entry data (current released subset)")
    #
    parser.add_argument("--config_path", default=None, help="Path to configuration options file")
    parser.add_argument("--config_name", default="site_info", help="Configuration section name")

    parser.add_argument("--input_file_list_path", default=None, help="Input file containing file paths to scan")
    parser.add_argument("--output_file_list_path", default=None, help="Output file containing file paths scanned")
    parser.add_argument("--fail_file_list_path", default=None, help="Output file containing file paths that fail scan")
    parser.add_argument("--scan_data_file_path", default=None, help="Output working file storing scan data (Pickle)")
    parser.add_argument("--coverage_file_path", default=None, help="Coverage map (JSON) output path")
    parser.add_argument("--type_map_file_path", default=None, help="Type map (JSON) output path")

    parser.add_argument("--num_proc", default=2, help="Number of processes to execute (default=2)")
    parser.add_argument("--chunk_size", default=10, help="Number of files loaded per process")
    parser.add_argument("--file_limit", default=None, help="Load file limit for testing")
    parser.add_argument("--debug", default=False, action='store_true', help="Turn on verbose logging")
    parser.add_argument("--mock", default=False, action='store_true', help="Use MOCK repository configuration for testing")
    parser.add_argument("--working_path", default=None, help="Working path for temporary files")
    args = parser.parse_args()
    #
    debugFlag = args.debug
    if debugFlag:
        logger.setLevel(logging.DEBUG)
    # ----------------------- - ----------------------- - ----------------------- - ----------------------- - ----------------------- -
    #                                       Configuration Details
    configPath = args.config_path
    configName = args.config_name
    if not configPath:
        configPath = os.getenv('DBLOAD_CONFIG_PATH', None)
    try:
        if os.access(configPath, os.R_OK):
            os.environ['DBLOAD_CONFIG_PATH'] = configPath
            logger.info("Using configuation path %s (%s)" % (configPath, configName))
        else:
            logger.error("Missing or access issue with config file %r" % configPath)
            exit(1)
        mockTopPath = os.path.join(TOPDIR, 'rcsb', 'mock-data', ) if args.mock else None
        cfgOb = ConfigUtil(configPath=configPath, defaultSectionName=configName, mockTopPath=mockTopPath)
    except Exception as e:
        logger.error("Missing or access issue with config file %r with %s" % (configPath, str(e)))
        exit(1)

    #
    try:
        numProc = int(args.num_proc)
        chunkSize = int(args.chunk_size)
        fileLimit = int(args.file_limit) if args.file_limit else None
        #
        failedFilePath = args.fail_file_list_path

        scanType = args.scanType
        #
        inputFileListPath = args.input_file_list_path
        outputFileListPath = args.output_file_list_path
        scanDataFilePath = args.scan_data_file_path
        dataCoverageFilePath = args.coverage_file_path
        dataTypeFilePath = args.type_map_file_path
        dictFilePath = args.dict_file_path if args.dict_file_path else os.path.join(TOPDIR, 'rcsb', 'mock-data', 'dictionaries', 'mmcif_pdbx_v5_next.dic')
        workPath = args.working_path if args.working_path else '.'
    except Exception as e:
        logger.exception("Argument processing problem %s" % str(e))
        parser.print_help(sys.stderr)
        exit(1)
    # ----------------------- - ----------------------- - ----------------------- - ----------------------- - ----------------------- -
    #
    # Read any input path lists -
    #
    inputPathList = None
    if inputFileListPath:
        mu = MarshalUtil(workPath=workPath)
        inputPathList = mu.doImport(inputFileListPath, format='list')
    #
    ##

    if args.scan_chem_comp_ref:
        contentType = 'chem_comp'

    elif args.scan_chem_comp_core_ref:
        contentType = 'chem_comp_core'

    elif args.scan_bird_chem_comp_ref:
        contentType = 'bird_chem_comp'

    elif args.scan_bird_chem_comp_core_ref:
        contentType = 'bird_chem_comp_core'

    elif args.scan_bird_ref:
        contentType = 'bird'

    elif args.scan_bird_family_ref:
        contentType = 'bird_family'

    elif args.scan_entry_data:
        contentType = 'pdbx'

    ok = scanRepo(cfgOb, contentType, scanDataFilePath, dictFilePath, numProc, chunkSize, fileLimit, scanType=scanType,
                  inputPathList=inputPathList, pathListFilePath=outputFileListPath, dataCoverageFilePath=dataCoverageFilePath,
                  dataTypeFilePath=dataTypeFilePath, failedFilePath=failedFilePath, workPath=workPath)

    logger.info("Operation completed with status %r " % ok)


if __name__ == '__main__':
    main()
